import re
import logging

import numpy as np
import pandas as pd

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.SequenceParams import Chain
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.reference_encoding.MatchedRegexEncoder import MatchedRegexEncoder
from immuneML.util.Logger import print_log
from immuneML.util.ReadsType import ReadsType


class MatchedRegexRepertoireEncoder(MatchedRegexEncoder):

    def _encode_new_dataset(self, dataset, params: EncoderParams):
        self._load_regex_df()

        feature_annotations = self._get_feature_info()
        encoded_repertoires, labels = self._encode_repertoires(dataset, params)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(
            examples=encoded_repertoires,
            example_ids=dataset.get_example_ids(),
            feature_names=list(feature_annotations["locus_id"]),
            feature_annotations=feature_annotations,
            labels=labels,
            encoding=MatchedRegexEncoder.__name__,
            info={'sequence_type': params.sequence_type,
                  'region_type': params.region_type}
        )

        return encoded_dataset

    def _get_feature_info(self):
        """
        returns a pandas dataframe containing:
         - feature id (id_CHAIN)
         - regex
         - v_gene (if match_v_genes == True)
        only for the motifs for which a regex was specified
        """
        features = {"regex_id": [], "locus_id": [], "locus": [], "regex": []}

        if self.match_v_genes:
            features["v_call"] = []

        for index, row in self.regex_df.iterrows():
            for chain_type in self.chains:
                regex = row[f"{chain_type}_regex"]

                if regex is not None:
                    features["regex_id"].append(f"{row['id']}")
                    features["locus_id"].append(f"{row['id']}_{chain_type}")
                    features["locus"].append(Chain.get_chain(chain_type).name.lower())
                    features["regex"].append(regex)

                    if self.match_v_genes:
                        v_gene = row[f"{chain_type}V"] if f"{chain_type}V" in row else None
                        features["v_call"].append(v_gene)

        return pd.DataFrame(features)

    def _encode_repertoires(self, dataset: RepertoireDataset, params: EncoderParams):
        # Rows = repertoires, Columns = regex matches (one chain per column)
        encoded_repertoires = np.zeros((dataset.get_example_count(), self.feature_count), dtype=int)
        labels = {label: [] for label in params.label_config.get_labels_by_name()} if params.encode_labels else None

        n_repertoires = dataset.get_example_count()

        for i, repertoire in enumerate(dataset.get_data()):
            print_log(f"Encoding repertoire {i+1}/{n_repertoires}", include_datetime=True)
            encoded_repertoires[i] = self._match_repertoire_to_regexes(repertoire, params)

            if labels is not None:
                for label_name in params.label_config.get_labels_by_name():
                    labels[label_name].append(repertoire.metadata[label_name])

        return encoded_repertoires, labels

    def _match_repertoire_to_regexes(self, repertoire: Repertoire, params: EncoderParams):
        matches = np.zeros(self.feature_count, dtype=int)
        rep_seqs = repertoire.sequences(params.region_type)

        match_idx = 0

        for index, row in self.regex_df.iterrows():
            for chain_type in self.chains:
                regex = row[f"{chain_type}_regex"]

                if regex is not None:
                    v_gene = row[f"{chain_type}V"] if f"{chain_type}V" in row else None

                    for rep_seq in rep_seqs:
                        if rep_seq.locus is not None:
                            if rep_seq.locus == chain_type:
                                if self._matches(rep_seq, regex, v_gene):
                                    n_matches = 1 if self.reads == ReadsType.UNIQUE else rep_seq.duplicate_count
                                    if n_matches is None:
                                        logging.warning(f"MatchedRegexRepertoireEncoder: count not defined for sequence with id {rep_seq.sequence_id} in repertoire {repertoire.identifier}, ignoring sequence...")
                                        n_matches = 0
                                    matches[match_idx] += n_matches
                        else:
                            logging.warning(f"{MatchedRegexRepertoireEncoder.__class__.__name__}: chain was not set for sequence {rep_seq.sequence_id}, skipping the sequence for matching...")
                    match_idx += 1

        return matches

    def _matches(self, receptor_sequence, regex, v_gene=None):
        if v_gene is not None and receptor_sequence.v_call != v_gene:
            matches = False
        else:
            matches = bool(re.search(regex, receptor_sequence.sequence_aa))

        return matches



import logging

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.util.ParameterValidator import ParameterValidator


class LabelHelper:

    @staticmethod
    def check_label_format(labels: list, instruction_name: str, yaml_location: str):
        ParameterValidator.assert_type_and_value(labels, list, instruction_name, f'{yaml_location}/labels')

        assert all(isinstance(label, str) or isinstance(label, dict) for label in labels), \
            f"{instruction_name}: labels under {yaml_location} were not defined properly. The list of labels has to either be a list of " \
            f"label names, or there can be a parameter 'positive_class' defined under the label name, for example:\n" \
            f"labels: # one label with no positive class (T1D) and one with a positive class (CMV)\n" \
            f"- T1D\n" \
            f"- CMV: # when defining a positive class, make sure to use the correct indentation\n" \
            f"    positive_class: True\n" \

        assert all(len(list(label.keys())) == 1 and isinstance(list(label.values())[0], dict) and 'positive_class' in list(label.values())[0]
                   and len(list(list(label.values())[0].keys())) == 1 for label in [l for l in labels if isinstance(l, dict)]), \
            f"{instruction_name}: The only legal parameter under a label name is 'positive_class'. If 'positive_class' is not specified, please " \
            f"remove the colon after the label name. "

    @staticmethod
    def create_label_config(labels: list, dataset: Dataset, instruction_name: str, yaml_location: str) -> LabelConfiguration:
        LabelHelper.check_label_format(labels, instruction_name, yaml_location)

        label_config = LabelConfiguration()
        for label in labels:
            label_name = label if isinstance(label, str) else list(label.keys())[0]
            positive_class = label[label_name]['positive_class'] if isinstance(label, dict) else None
            if dataset.labels is not None and label_name in dataset.labels:
                label_values = list(dataset.labels[label_name])
            elif hasattr(dataset, "get_metadata"):
                label_values = list(set(dataset.get_metadata([label_name])[label_name]))
            else:
                label_values = []
                logging.warning(f"{instruction_name}: for {yaml_location}, label values could not be recovered for label "
                              f"{label}, using empty list instead. This issue may occur due to improper loading of dataset {dataset.name},"
                              f"and could cause problems with some encodings.")

            label_config.add_label(label_name, label_values, positive_class=positive_class)

        return label_config



from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod, get_data_for_clustering
from sklearn.cluster import AgglomerativeClustering as SklearnAgglomerativeClustering


class AgglomerativeClustering(ClusteringMethod):
    """
        Agglomerative clustering method which wraps scikit-learn's clustering of the same name.
        Input arguments for the method are the same as supported by scikit-learn (see `AgglomerativeClustering scikit-learn documentation
        <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.AgglomerativeClustering.html>`_ for details).

        **YAML specification:**

        .. indent with spaces
        .. code-block:: yaml

            definitions:
                ml_methods:
                    my_agglomerative_clustering:
                        AgglomerativeClustering:
                            # arguments as defined by scikit-learn
                            n_clusters: 3
                            linkage: 'ward'
        """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.model = SklearnAgglomerativeClustering(**kwargs)

    def fit(self, dataset: Dataset):
        data = get_data_for_clustering(dataset)
        self.model.fit(data)

    def fit_predict(self, dataset: Dataset):
        data = get_data_for_clustering(dataset)
        return self.model.fit_predict(data)

    def predict(self, dataset: Dataset):
        raise RuntimeError("Agglomerative Clustering does not support predict method")

    def transform(self, dataset: Dataset):
        raise RuntimeError("Agglomerative Clustering does not support transform method")



import argparse
import sys
from pathlib import Path

from immuneML.api.galaxy.Util import Util
from immuneML.data_model.bnp_util import write_yaml
from immuneML.util.PathBuilder import PathBuilder


def parse_command_line_arguments(args):
    parser = argparse.ArgumentParser(description="Tool for building specification for training generative "
                                                 "models in Galaxy")

    parser.add_argument("-e", "--gen_example_count", required=True,
                        help="Number of examples to generate.")
    parser.add_argument("-c", "--chain_type", required=True, choices=["TRA", "TRB", "IGH", "IGK", "IGL", "TRD", "TRG"],
                        help="Chain type of the sequences to generate.")
    parser.add_argument("-m", "--generative_method", choices=["SoNNia", "SimpleLSTM", "PWM", "SimpleVAE"],
                        required=True, help="Which generative model should be trained.")

    parser.add_argument("-s", "--sequence_length_report", choices=["True", "False"], default="False",
                        help="Whether to run the SequenceLengthDistribution report.")
    parser.add_argument("-q", "--amino_acid_report", choices=["True", "False"], default="False",
                        help="Whether to run the AminoAcidFrequencyDistribution report.")
    parser.add_argument("-k", "--kl_gen_model_report", choices=["True", "False"], default="False",
                        help="Whether to run the KLKmerComparison report.")

    parser.add_argument("-t", "--training_percentage", type=float, required=True,
                        help="The percentage of data used for training.")
    parser.add_argument("-x", "--export_dataset_type", choices=["generated_dataset", "combined_dataset"],
                        default="generated_dataset",
                        help="Whether to export only the generated dataset, or the combined training+(test+)generated dataset.")

    parser.add_argument("-o", "--output_path", required=True,
                        help="Output location for the generated yaml file (directory).")
    parser.add_argument("-f", "--file_name", default="specs.yaml",
                        help="Output file name for the yaml file. Default name is 'specs.yaml' if not specified.")

    return parser.parse_args(args)


def build_specs(parsed_args):
    reports = {}
    if parsed_args.sequence_length_report == "True":
        reports['sequence_length_distribution'] = "SequenceLengthDistribution"
    if parsed_args.amino_acid_report == "True":
        reports['amino_acid_frequency_distribution'] = "AminoAcidFrequencyDistribution"
    if parsed_args.kl_gen_model_report == "True":
        reports['kl_kmer_comparison'] = "KLKmerComparison"

    gen_model_args = {"locus": parsed_args.chain_type}

    if parsed_args.generative_method == "SimpleVAE":
        reports['generative_model_overview'] = "VAESummary"
        gen_model_args['region_type'] = 'imgt_cdr3'
    elif parsed_args.generative_method == "PWM":
        reports['generative_model_overview'] = "PWMSummary"
    elif parsed_args.generative_method == "SimpleLSTM":
        gen_model_args['region_type'] = 'imgt_cdr3'

    if parsed_args.generative_method == "SoNNia":
        gen_model_args["default_model_name"] = f"human{parsed_args.chain_type}"

    specs = {
        "definitions": {
            "datasets": {
                "dataset": {
                    "format": "AIRR",
                    "params": {"dataset_file": Util.discover_dataset_path()}
                }
            },
            "reports": reports,
            "ml_methods": {
                "generative_model": {
                    parsed_args.generative_method: gen_model_args
                }
            }
        },
        "instructions": {
            f"train_{parsed_args.generative_method}": {
                "type": "TrainGenModel",
                "dataset": "dataset",
                "method": "generative_model",
                "number_of_processes": 8,
                "gen_examples_count": int(parsed_args.gen_example_count),
                "training_percentage": float(parsed_args.training_percentage) / 100,
                "export_generated_dataset": True if parsed_args.export_dataset_type == "generated_dataset" else False,
                "export_combined_dataset": True if parsed_args.export_dataset_type == "combined_dataset" else False,
                "reports": list(reports.keys()),
            }
        }
    }

    return specs


def main(args):
    parsed_args = parse_command_line_arguments(args)
    specs = build_specs(parsed_args)

    PathBuilder.build(parsed_args.output_path)
    output_location = Path(parsed_args.output_path) / parsed_args.file_name

    write_yaml(output_location, specs)

    return str(output_location)


if __name__ == "__main__":
    main(sys.argv[1:])



from dataclasses import dataclass
from pathlib import Path

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.hyperparameter_optimization.HPSetting import HPSetting


@dataclass
class MLApplicationState:

    dataset: Dataset
    hp_setting: HPSetting
    label_config: LabelConfiguration
    pool_size: int
    name: str
    metrics: list = None
    path: Path = None
    predictions_path: Path = None
    metrics_path: Path = None



import logging
from pathlib import Path

import pandas as pd
import plotly.express as px

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.encoding_reports.FeatureReport import FeatureReport
from immuneML.util.ParameterValidator import ParameterValidator


class FeatureComparison(FeatureReport):
    """
    Encoding a dataset results in a numeric matrix, where the rows are examples (e.g., sequences, receptors, repertoires)
    and the columns are features. For example, when :ref:`KmerFrequency` encoder is used, the features are the
    k-mers (AAA, AAC, etc..) and the feature values are the frequencies per k-mer.

    This report separates the examples based on a binary metadata label, and plots the mean feature value
    of each feature in one example group against the other example group (for example: plot the feature
    value of 'sick' repertoires on the x axis, and 'healthy' repertoires on the y axis to spot consistent differences).
    The plot can be separated into different colors or facets using other metadata labels
    (for example: plot the average feature values of 'cohort1', 'cohort2' and 'cohort3' in different colors to spot biases).

    Alternatively, when plotting features without comparing them across a binary label, see:
    :py:obj:`~immuneML.reports.encoding_reports.FeatureValueBarplot.FeatureValueBarplot` report to plot
    a simple bar chart per feature (average across examples).
    Or :py:obj:`~immuneML.reports.encoding_reports.FeatureDistribution.FeatureDistribution` report to plot
    the distribution of each feature across examples, rather than only showing the mean value in a bar plot.


    Example output:

    .. image:: ../../_static/images/reports/feature_comparison_zoom.png
       :alt: Feature comparison zoomed in plot with VLEQ highlighted
       :width: 650



    **Specification arguments:**

    - comparison_label (str): Mandatory label. This label is used to split the encoded data matrix and define the x and y axes of the plot.
      This label is only allowed to have 2 classes (for example: sick and healthy, binding and non-binding).

    - color_grouping_label (str): Optional label that is used to color the points in the scatterplot. This can not be the same as comparison_label.

    - row_grouping_label (str): Optional label that is used to group scatterplots into different row facets. This can not be the same as comparison_label.

    - column_grouping_label (str): Optional label that is used to group scatterplots into different column facets. This can not be the same as comparison_label.

    - show_error_bar (bool): Whether to show the error bar (standard deviation) for the points, both in the x and y dimension.

    - log_scale (bool): Whether to plot the x and y axes in log10 scale (log_scale = True) or continuous scale (log_scale = False). By default, log_scale is False.

    - keep_fraction (float): The total number of features may be very large and only the features differing significantly across
      comparison labels may be of interest. When the keep_fraction parameter is set below 1, only the fraction of features that
      differs the most across comparison labels is kept for plotting (note that the produced .csv file still contains all data).
      By default, keep_fraction is 1, meaning that all features are plotted.

    - opacity (float): a value between 0 and 1 setting the opacity for data points making it easier to see if there are overlapping points


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_comparison_report:
                    FeatureComparison: # compare the different classes defined in the label disease
                        comparison_label: disease

    """

    @classmethod
    def build_object(cls, **kwargs):
        comparison_label = kwargs["comparison_label"] if "comparison_label" in kwargs else None
        color_grouping_label = kwargs["color_grouping_label"] if "color_grouping_label" in kwargs else None
        row_grouping_label = kwargs["row_grouping_label"] if "row_grouping_label" in kwargs else None
        column_grouping_label = kwargs["column_grouping_label"] if "column_grouping_label" in kwargs else None
        log_scale = kwargs["log_scale"] if "log_scale" in kwargs else None
        keep_fraction = float(kwargs["keep_fraction"]) if "keep_fraction" in kwargs else 1.0
        ParameterValidator.assert_type_and_value(keep_fraction, float, "FeatureComparison", "keep_fraction", min_inclusive=0, max_inclusive=1)
        ParameterValidator.assert_type_and_value(log_scale, bool, "FeatureComparison", "log_scale")

        assert comparison_label is not None, "FeatureComparison: the parameter 'comparison_label' must be set in order to be able to compare across this label"

        assert comparison_label != color_grouping_label, f"FeatureComparison: comparison label {comparison_label} can not be used as color_grouping_label"
        assert comparison_label != row_grouping_label, f"FeatureComparison: comparison label {comparison_label} can not be used as row_grouping_label"
        assert comparison_label != column_grouping_label, f"FeatureComparison: comparison label {comparison_label} can not be used as column_grouping_label"

        return FeatureComparison(**kwargs)

    def __init__(self, dataset: Dataset = None, result_path: Path = None, comparison_label: str = None,
                 color_grouping_label: str = None, row_grouping_label=None, column_grouping_label=None, opacity: float = 0.7,
                 show_error_bar=True, log_scale: bool = False, keep_fraction: int = 1, number_of_processes: int = 1, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, color_grouping_label=color_grouping_label,
                         row_grouping_label=row_grouping_label, column_grouping_label=column_grouping_label,
                         number_of_processes=number_of_processes, name=name)
        self.comparison_label = comparison_label
        self.show_error_bar = show_error_bar
        self.log_scale = log_scale
        self.keep_fraction = keep_fraction
        self.opacity = opacity
        self.result_name = "feature_comparison"
        self.name = name

    def _generate(self):
        result = self._generate_report_result()
        result.info = ("Compares the feature values in a given encoded data matrix across two values for a metadata "
                       "label. Each point in the resulting scatterplot represents one feature, and the values on the "
                       "x and y axes are the average feature values across examples of two different classes. ")
        return result

    def _plot(self, data_long_format) -> ReportOutput:
        groupby_cols = [self.comparison_label, self.x, self.color, self.facet_row, self.facet_column]
        groupby_cols = [i for i in groupby_cols if i]
        groupby_cols = list(set(groupby_cols))
        plotting_data = data_long_format.groupby(groupby_cols, as_index=False).agg(
            {"value": ['mean', self.std]})

        plotting_data.columns = plotting_data.columns.map(''.join)

        unique_label_values = plotting_data[self.comparison_label].unique()
        assert len(unique_label_values) == 2, \
            f"FeatureComparison: comparison label {self.comparison_label} does not have 2 values; {unique_label_values}"
        class_x, class_y = unique_label_values

        merge_labels = [label for label in ["feature", self.color, self.facet_row, self.facet_column] if label]

        plotting_data = pd.merge(plotting_data.loc[plotting_data[self.comparison_label] == class_x],
                                 plotting_data.loc[plotting_data[self.comparison_label] == class_y],
                                 on=merge_labels)

        if plotting_data.shape[0] == 0:
            logging.warning(f"{FeatureComparison.__name__}: there is no overlap between the (combination of) values of {merge_labels} for "
                            f"different values the {self.comparison_label}.")

        plotting_data = self._filter_keep_fraction(plotting_data) if self.keep_fraction < 1 else plotting_data

        error_x = "valuestd_x" if self.show_error_bar else None
        error_y = "valuestd_y" if self.show_error_bar else None

        figure = px.scatter(plotting_data, x="valuemean_x", y="valuemean_y", error_x=error_x, error_y=error_y,
                            color=self.color, facet_row=self.facet_row, facet_col=self.facet_column, hover_name="feature",
                            log_x=self.log_scale, log_y=self.log_scale, opacity=self.opacity,
                            labels={
                                "valuemean_x": f"Average feature values for {self.comparison_label} = {class_x}",
                                "valuemean_y": f"Average feature values for {self.comparison_label} = {class_y}",
                            }, template='plotly_white',
                            color_discrete_sequence=px.colors.diverging.Tealrose)

        self.add_diagonal(figure, x_range=(plotting_data['valuemean_x'].min(), plotting_data["valuemean_x"].max()),
                          y_range=(plotting_data['valuemean_y'].min(), plotting_data["valuemean_y"].max()))

        file_path = self.result_path / f"{self.result_name}.html"

        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name=f"Comparison of feature values across {self.comparison_label}")

    def add_diagonal(self, figure, x_range: tuple, y_range: tuple):
        figure.update_layout(shapes=[{'type': "line", 'line': dict(color="#B0C2C7", dash="dash"),
                                      'y0': min(x_range[0], y_range[0]), 'y1': max(x_range[1], y_range[1]),
                                      'x0': min(x_range[0], y_range[0]), 'x1': max(x_range[1], y_range[1]),
                                      'layer': 'below'}])

    def _filter_keep_fraction(self, plotting_data):
        plotting_data["diff_xy"] = abs(plotting_data["valuemean_x"] - plotting_data["valuemean_y"])
        plotting_data.sort_values(by="diff_xy", inplace=True, ascending=False)
        plotting_data.drop(columns="diff_xy", inplace=True)

        keep_nrows = round(plotting_data.shape[0] * self.keep_fraction)
        return plotting_data.head(keep_nrows)

    def check_prerequisites(self):
        location = self.__class__.__name__
        run_report = True

        if self.dataset.encoded_data is None or self.dataset.encoded_data.examples is None:
            logging.warning(
                f"{location}: this report can only be created for an encoded dataset. {location} report will not be created.")
            run_report = False
        elif len(self.dataset.encoded_data.examples.shape) != 2:
            logging.warning(
                f"{location}: this report can only be created for a 2-dimensional encoded dataset. {location} report will not be created.")
            run_report = False
        else:
            legal_labels = list(self.dataset.get_label_names())

            if self.comparison_label not in legal_labels:
                logging.warning(
                    f"{location}: comparison_label was not defined. {location} report will not be created.")
                run_report = False
            elif len(set(self.dataset.get_metadata([self.comparison_label])[self.comparison_label])) != 2:
                logging.warning(
                    f"{location}: comparison label {self.comparison_label} does not have 2 values: {set(self.dataset.get_metadata([self.comparison_label])[self.comparison_label])}. {location} report will not be created.")
                run_report = False
            else:
                legal_labels.remove(self.comparison_label)

                for label_param in [self.color, self.facet_row, self.facet_column]:
                    if label_param is not None:
                        if label_param == self.comparison_label:
                            logging.warning(
                                f"{location}: comparison label '{self.comparison_label}' can not be used in other fields. {location} report will not be created.")
                            run_report = False
                        if label_param not in legal_labels:
                            logging.warning(
                                f"{location}: undefined label '{label_param}'. Legal options are: {legal_labels}. {location} report will not be created.")
                            run_report = False

        return run_report



from sklearn.ensemble import RandomForestClassifier as RFC

from immuneML.ml_methods.classifiers.SklearnMethod import SklearnMethod
from scripts.specification_util import update_docs_per_mapping


class RandomForestClassifier(SklearnMethod):
    """
    This is a wrapper of scikit-learn’s RandomForestClassifier class. Please see the
    `RandomForestClassifier scikit-learn documentation <https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html>`_
    of RandomForestClassifier for the parameters.

    Note: if you are interested in plotting the coefficients of the random forest classifier model,
    consider running the :ref:`Coefficients` report.

    For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_random_forest_classifier: # user-defined method name
                    RandomForestClassifier: # name of the ML method
                        # sklearn parameters (same names as in original sklearn class)
                        random_state: 100 # always use this value for random state
                        n_estimators: [10, 50, 100] # find the optimal number of trees in the forest
                        # Additional parameter that determines whether to print convergence warnings
                        show_warnings: True
                    # if any of the parameters under RandomForestClassifier is a list and model_selection_cv is True,
                    # a grid search will be done over the given parameters, using the number of folds specified in model_selection_n_folds,
                    # and the optimal model will be selected
                    model_selection_cv: True
                    model_selection_n_folds: 5
                # alternative way to define ML method with default values:
                my_default_random_forest: RandomForestClassifier

    """

    def __init__(self, parameter_grid: dict = None, parameters: dict = None):
        parameters = parameters if parameters is not None else {}

        if parameter_grid is not None:
            parameter_grid = parameter_grid
        else:
            parameter_grid = {"n_estimators": [10, 50, 100]}

        super(RandomForestClassifier, self).__init__(parameter_grid=parameter_grid, parameters=parameters)

    def _get_ml_model(self, cores_for_training: int = 2, X=None):
        params = self._parameters.copy()
        params["n_jobs"] = cores_for_training
        return RFC(**params)

    def can_predict_proba(self) -> bool:
        return True

    def can_fit_with_example_weights(self) -> bool:
        return True

    def get_params(self):
        params = self.model.get_params(deep=True)
        params["feature_importances"] = self.model.feature_importances_.tolist()
        return params

    @staticmethod
    def get_documentation():
        doc = str(RandomForestClassifier.__doc__)

        mapping = {
            "For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.": SklearnMethod.get_usage_documentation("RandomForestClassifier"),
        }

        doc = update_docs_per_mapping(doc, mapping)
        return doc



import torch
from torch import nn


class SimpleLSTMGenerator(nn.Module):

    def __init__(self, input_size, embed_size, hidden_size, output_size, batch_size, num_layers=1, device: str = 'cpu'):
        super(SimpleLSTMGenerator, self).__init__()

        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.batch_size = batch_size
        self.device = device

        self.embed = nn.Embedding(num_embeddings=input_size, embedding_dim=embed_size)
        nn.init.normal_(self.embed.weight)

        self.lstm = nn.LSTM(input_size=embed_size, hidden_size=hidden_size, num_layers=num_layers)
        for name, param in self.lstm.named_parameters():
            if 'weight' in name:
                nn.init.xavier_uniform_(param)
            elif 'bias' in name:
                nn.init.zeros_(param)

        self.fc = nn.Linear(hidden_size, output_size)
        nn.init.xavier_uniform_(self.fc.weight)
        nn.init.zeros_(self.fc.bias)

    def forward(self, features, hidden_and_cell_state):
        # features shape: (batch_size, seq_len)
        features = features.transpose(0, 1)  # Convert to (seq_len, batch_size)
        embedded = self.embed(features)  # (seq_len, batch_size, embed_size)
        
        output, hidden_and_cell_state = self.lstm(embedded, hidden_and_cell_state)
        # output shape: (seq_len, batch_size, hidden_size)
        
        output = self.fc(output)  # (seq_len, batch_size, output_size)
        output = output.transpose(0, 1)  # Convert back to (batch_size, seq_len, output_size)
        return output, hidden_and_cell_state

    def init_zero_state(self, batch_size=None):
        if batch_size is None:
            batch_size = self.batch_size
        init_hidden = torch.zeros(self.num_layers, batch_size, self.hidden_size, device=self.device)
        init_cell = torch.zeros(self.num_layers, batch_size, self.hidden_size, device=self.device)
        return init_hidden, init_cell



import argparse
import sys
from pathlib import Path

from immuneML.data_model.bnp_util import write_yaml
from immuneML.util.PathBuilder import PathBuilder


def parse_command_line_arguments(args):
    parser = argparse.ArgumentParser(description="Tool for building specification for simulating data with LiGO in Galaxy.")

    parser.add_argument("-m", "--motif_seed", required=True,
                        help="Seed of the motif to implant.")
    parser.add_argument("-e", "--example_with_motif_count", required=True, type=int,
                        help="Number of examples to generate containing the motif.")
    parser.add_argument("-w", "--example_without_motif_count", required=True, type=int,
                        help="Number of examples to generate NOT containing the motif.")
    parser.add_argument("-s", "--simulation_strategy",  choices=["RejectionSampling", "Implanting"], required=True,
                        help="Strategy for simulating signals into sequences.")
    parser.add_argument("-d", "--dataset_type",  choices=["repertoire", "sequence"], required=True,
                        help="Type of dataset to generate (RepertoireDataset or SequenceDataset, ReceptorDataset is currently not available)")
    parser.add_argument("-c", "--chain_type",  choices=["humanTRA", "humanTRB", "humanIGH", "humanIGK", "humanIGL"], required=True,
                        help="Chain type for the simulated dataset, to be used for default OLGA model selection.")
    parser.add_argument("-p", "--signal_percentage",  type=float, default=100,
                        help="Percentage of sequences that contain the signal.")
    parser.add_argument("-r", "--repertoire_size",  type=int, required=False,
                        help="Number of sequences per repertoire, if dataset_type is repertoire.")

    parser.add_argument("-o", "--output_path", required=True,
                        help="Output location for the generated yaml file (directory).")
    parser.add_argument("-f", "--file_name", default="specs.yaml",
                        help="Output file name for the yaml file. Default name is 'specs.yaml' if not specified.")

    return parser.parse_args(args)


def build_specs(parsed_args):

    specs = {
        "definitions": {
            "motifs": {
                "motif1": {
                    "seed": parsed_args.motif_seed
                }
            },
            "signals": {
                "signal1": {
                    "motifs": ["motif1"]
                }
            },
            "simulations": {
                "sim1": {
                    "sim_items": {
                        "signal": {
                            "generative_model": {
                                "default_model_name": parsed_args.chain_type,
                                "type": "OLGA"
                            },
                            "number_of_examples": parsed_args.example_with_motif_count,
                            "signals": {"signal1": parsed_args.signal_percentage / 100},
                            "seed": 100,
                            "receptors_in_repertoire_count": parsed_args.repertoire_size if parsed_args.dataset_type == "repertoire" else None
                        },
                        "no_signal": {
                            "generative_model": {
                                "default_model_name": parsed_args.chain_type,
                                "type": "OLGA"
                            },
                            "number_of_examples": parsed_args.example_without_motif_count,
                            "signals": {},
                            "seed": 200,
                            "receptors_in_repertoire_count": parsed_args.repertoire_size if parsed_args.dataset_type == "repertoire" else None
                        }
                    },
                    "is_repertoire": True if parsed_args.dataset_type == "repertoire" else False,
                    "sequence_type": "amino_acid",
                    "simulation_strategy": parsed_args.simulation_strategy
                }
            }
        },
        "instructions": {
            f"simulate_with_ligo": {
                "type": "LigoSim",
                "simulation": "sim1",
                "number_of_processes": 8,
                "export_p_gens": False,
            }
        }
    }

    return specs

def main(args):
    parsed_args = parse_command_line_arguments(args)
    specs = build_specs(parsed_args)

    PathBuilder.build(parsed_args.output_path)
    output_location = Path(parsed_args.output_path) / parsed_args.file_name

    write_yaml(output_location, specs)

    return str(output_location)


if __name__ == "__main__":
    main(sys.argv[1:])



import logging

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from sklearn.metrics import roc_curve

from immuneML.environment.Constants import Constants
from immuneML.environment.Label import Label
from immuneML.hyperparameter_optimization.states.HPItem import HPItem
from immuneML.ml_metrics.ml_metrics import roc_auc_score
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.train_ml_model_reports.TrainMLModelReport import TrainMLModelReport
from immuneML.util.PathBuilder import PathBuilder


class ROCCurveSummary(TrainMLModelReport):
    """
    This report plots ROC curves for all trained ML settings ([preprocessing], encoding, ML model) in the outer loop of cross-validation in the
    :ref:`TrainMLModel` instruction. If there are multiple splits in the outer loop, this report will make one plot per split. This report is
    defined only for binary classification. If there are multiple labels defined in the instruction, each label has to have two classes to be included
    in this report.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_roc_summary_report: ROCCurveSummary

    """

    @classmethod
    def build_object(cls, **kwargs):
        return ROCCurveSummary(**kwargs)

    def _generate(self) -> ReportResult:
        report_result = ReportResult(name=self.name,
                                     info="Plots ROC curves for all trained ML settings ([preprocessing], encoding, "
                                          "ML model) in the outer loop of cross-validation in the TrainMLModel "
                                          "instruction")

        PathBuilder.build(self.result_path)

        for label in self.state.label_configuration.get_label_objects():
            if len(label.values) != 2:
                logging.warning(f"{ROCCurveSummary.__name__}: report {self.name} is skipping label {label.name} "
                                f"as it has {len(label.values)} classes, while this report expects 2 classes.")
            elif label.positive_class is None:
                logging.warning(f"{ROCCurveSummary.__name__}: report {self.name} is skipping label {label.name} "
                                f"because 'positive_class' parameter is not set.")
            else:
                for index in range(self.state.assessment.split_count):
                    figure = self._create_figure_for_assessment_split(index, label)
                    report_result.output_figures.append(figure)

        return report_result

    def _create_figure_for_assessment_split(self, index, label: Label):
        data = []
        for hp_item_name, hp_item in self.state.assessment_states[index].label_states[label.name].assessment_items.items():
            data.append(self._make_roc_curve(hp_item, label, f"{label.name}_{label.positive_class}_proba"))

        figure = self._draw_rocs(data=data, roc_legends=[f"{item['HPItem']} (AUROC = {round(item['AUC'], 2)})" for item in data],
                                 figure_name=f"ROC curves for label {label.name} on assessment split {index + 1}.html")
        return figure

    def _make_roc_curve(self, hp_item: HPItem, label: Label, proba_name: str) -> dict:
        df = pd.read_csv(hp_item.test_predictions_path)

        true_y = df[f"{label.name}_true_class"].values

        if hp_item.method.can_predict_proba():
            predicted_y = df[proba_name].values
        else:
            predicted_y = df[f"{label.name}_predicted_class"].values

        fpr, tpr, _ = roc_curve(y_true=true_y, y_score=predicted_y, pos_label=label.positive_class)

        return {
            "FPR": fpr,
            "TPR": tpr,
            "AUC": roc_auc_score(true_y=true_y, predicted_y=predicted_y),
            "HPItem": str(hp_item.hp_setting)
        }

    def _draw_rocs(self, data: list, roc_legends: list, figure_name: str) -> ReportOutput:
        figure = go.Figure()
        colors = px.colors.sequential.Viridis[::2][::-1]

        figure.add_trace(
            go.Scatter(x=[0, 1], y=[0, 1], mode='lines', name='baseline (AUC = 0.5)', line=dict(color=Constants.PLOTLY_BLACK, dash='dash'),
                       hoverinfo="skip"))

        for index, item in enumerate(data):
            figure.add_trace(
                go.Scatter(x=item["FPR"], y=item["TPR"], mode='lines', name=roc_legends[index], marker=dict(color=colors[index], line=dict(width=3)),
                           hoverinfo="skip"))

        figure.update_layout(template='plotly_white', xaxis_title='false positive rate', yaxis_title='true positive rate')
        figure.update_layout(legend=dict(yanchor="bottom", y=0.06, xanchor="right", x=0.99), font_size=15)

        file_path = self.result_path / figure_name.replace(" ", "_")
        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name=figure_name.split(".")[0])



from pathlib import Path
import numpy as np
import os

from immuneML.data_model.SequenceParams import RegionType
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder import CompAIRRSequenceAbundanceEncoder
from immuneML.encodings.abundance_encoding.KmerAbundanceEncoder import KmerAbundanceEncoder
from immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder import SequenceAbundanceEncoder
from immuneML.encodings.kmer_frequency.sequence_encoding.SequenceEncodingType import SequenceEncodingType
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.ParameterValidator import ParameterValidator


class SignificantFeaturesHelper:

    @staticmethod
    def parse_parameters(kwargs, location):
        ParameterValidator.assert_keys_present(kwargs.keys(), ["p_values", "k_values", "label", 'region_type', 'sequence_type'], location, location)

        ParameterValidator.assert_type_and_value(kwargs["p_values"], list, location, "p_values")
        ParameterValidator.assert_type_and_value(kwargs["k_values"], list, location, "k_values")

        assert isinstance(kwargs["label"], dict), f"{location}: {kwargs['label']} is not a valid value for parameter label. " \
                                                  f"It has to be of type dict, but is now of type {type(kwargs['label']).__name__}." \
                                                  f"Did you remember to set the positive_class?"

        assert len(kwargs["label"]) == 1, f"{location}: only one label is allowed to be set, found {len(kwargs['label'])}: {list(kwargs['label'])}"

        label_name = list(kwargs["label"].keys())[0]

        assert "positive_class" in kwargs["label"][label_name], f"{location}: positive_class must be set for label {label_name}"

        assert len(kwargs["p_values"]) == len(set( kwargs["p_values"])), f"{location}: p_values should only contain unique values, found {kwargs['p_values']}"
        assert len(kwargs["k_values"]) == len(set( kwargs["k_values"])), f"{location}: k_values should only contain unique values, found {kwargs['k_values']}"

        ParameterValidator.assert_all_type_and_value(kwargs["p_values"], float, "location", "p_values", min_inclusive=0)
        ParameterValidator.assert_region_type(kwargs, SignificantFeaturesHelper.__name__)
        ParameterValidator.assert_sequence_type(kwargs, SignificantFeaturesHelper.__name__)

        kwargs['region_type'] = RegionType[kwargs['region_type'].upper()]
        kwargs['sequence_type'] = SequenceType[kwargs['sequence_type'].upper()]

        for value in kwargs["k_values"]:
            if value != "full_sequence":
                ParameterValidator.assert_type_and_value(value, int, location, "k_values", 1)

        if "compairr_path" in kwargs and kwargs["compairr_path"] is not None:
            ParameterValidator.assert_type_and_value(kwargs["compairr_path"], str, location, "compairr_path")
            kwargs["compairr_path"] = Path(kwargs["compairr_path"])

        return kwargs

    @staticmethod
    def parse_sequences_path(kwargs, field_name, location):
        ParameterValidator.assert_keys_present(kwargs.keys(), [field_name], location, location)
        ParameterValidator.assert_type_and_value(kwargs[field_name], str, location,
                                                 field_name)
        assert os.path.isfile(kwargs[field_name]), f"{location}: implanted_sequences_path does not exist: {kwargs['field_name']}"

        kwargs[field_name] = Path(kwargs[field_name])

        return kwargs

    @staticmethod
    def load_sequences(ground_truth_sequences_path: Path):
        with open(ground_truth_sequences_path) as f:
            lines = f.readlines()
            sequences = [seq.strip() for seq in lines]
        return sequences

    @staticmethod
    def _get_encoder_name(k):
        encoder_name = f"{k}-mer" if type(k) == int else k
        return encoder_name

    @staticmethod
    def _build_encoder_params(label_config, encoder_result_path, region_type: RegionType, sequence_type: SequenceType):
        encoder_params = EncoderParams(result_path=encoder_result_path,
                                       label_config=label_config,
                                       pool_size=1,
                                       learn_model=True,
                                       encode_labels=False,
                                       region_type=region_type,
                                       sequence_type=sequence_type)

        return encoder_params

    @staticmethod
    def _build_kmer_encoder(dataset, k, p_value, encoder_params):
        encoder = KmerAbundanceEncoder(p_value_threshold=p_value,
                                       sequence_encoding=SequenceEncodingType.CONTINUOUS_KMER,
                                       k=k, k_left=0, k_right=0, min_gap=0, max_gap=0)

        encoder.encode(dataset, encoder_params)

        return encoder

    @staticmethod
    def _build_sequence_encoder(dataset, p_value, encoder_params: EncoderParams):
        encoder = SequenceAbundanceEncoder(comparison_attributes=[encoder_params.get_sequence_field_name()],
                                           p_value_threshold=p_value, sequence_batch_size=100000, repertoire_batch_size=16)

        encoder.encode(dataset, encoder_params)

        return encoder

    @staticmethod
    def _build_compairr_sequence_encoder(dataset, p_value, encoder_params, compairr_path):
        encoder = CompAIRRSequenceAbundanceEncoder(p_value_threshold=p_value, compairr_path=compairr_path,
                                                   sequence_batch_size=100000, ignore_genes=True, threads=8,
                                                   keep_temporary_files=True)

        encoder.encode(dataset, encoder_params)

        return encoder

    @staticmethod
    def _get_relevant_feature_presence(encoder, relevant_indices):

        if isinstance(encoder, KmerAbundanceEncoder):
            relevant_feature_presence = np.sum(encoder.kmer_presence_matrix[relevant_indices], axis=0)
        elif isinstance(encoder, CompAIRRSequenceAbundanceEncoder):
            relevant_feature_presence = np.sum(encoder.sequence_presence_matrix[relevant_indices], axis=0)
        else:
            relevant_feature_presence = np.zeros(shape=(6,))

            for i, sequence_vector in enumerate(encoder.comparison_data.get_item_vectors()):
                if relevant_indices[i]:
                    relevant_feature_presence += sequence_vector

        return relevant_feature_presence



import copy
import random
import uuid
from pathlib import Path

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.data_model.SequenceParams import RegionType, ChainPair
from immuneML.data_model.SequenceSet import ReceptorSequence, Receptor
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.datasets.ElementDataset import ReceptorDataset, SequenceDataset
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.RepertoireBuilder import RepertoireBuilder


class RandomDatasetGenerator:

    @staticmethod
    def _check_probabilities(probabilities_dict, key_type, dict_name):
        assert isinstance(probabilities_dict, dict) and all(
            isinstance(key, key_type) for key in probabilities_dict.keys()) \
               and all(isinstance(value, float) or value in {0, 1} for value in probabilities_dict.values()) \
               and 0.99 <= round(sum(probabilities_dict.values()), 5) <= 1, \
            f"RandomDatasetGenerator: {dict_name} are not specified correctly. They should be a dictionary with probabilities per count " \
            f"and sum to 1, but got {probabilities_dict} instead."

    @staticmethod
    def _check_labels(labels: dict):
        if labels is not None:
            assert isinstance(labels, dict)
            for label in labels:
                RandomDatasetGenerator._check_probabilities(labels[label], object, f"labels - {label}")

    @staticmethod
    def _check_example_count(count, name):
        assert isinstance(count, int) and count > 0, f"RandomDatasetGenerator: {name} is not specified properly. " \
                                                     f"It should be a positive integer, got {count} instead."

    @staticmethod
    def _check_path(path: Path):
        assert path is not None, "RandomDatasetGenerator: path cannot be None when generating datasets."

    @staticmethod
    def _check_rep_dataset_generation_params(repertoire_count: int, sequence_count_probabilities: dict,
                                             sequence_length_probabilities: dict,
                                             labels: dict, path: Path):

        RandomDatasetGenerator._check_example_count(repertoire_count, "repertoire_count")
        RandomDatasetGenerator._check_probabilities(sequence_count_probabilities, int, "sequence_count_probabilities")
        RandomDatasetGenerator._check_probabilities(sequence_length_probabilities, int, "sequence_length_probabilities")
        RandomDatasetGenerator._check_labels(labels)
        RandomDatasetGenerator._check_path(path)

    @staticmethod
    def generate_repertoire_dataset(repertoire_count: int, sequence_count_probabilities: dict,
                                    sequence_length_probabilities: dict,
                                    labels: dict, path: Path, name="repertoire_dataset") -> RepertoireDataset:
        """
        Creates repertoire_count repertoires where the number of sequences per repertoire is sampled from the probability distribution given
        in sequence_count_probabilities. The length of sequences is sampled independently for each sequence from
        sequence_length_probabilities distribution. The labels are also randomly assigned to repertoires from the distribution given in
        labels. In this case, labels are multi-class, so each repertoire will get at one class from each label. This means that negative
        classes for the labels should be included as well in the specification.

        An example of input parameters is given below:
        repertoire_count: 100 # generate 100 repertoires
        sequence_count_probabilities:
            100: 0.5 # half of the generated repertoires will have 100 sequences
            200: 0.5 # the other half of the generated repertoires will have 200 sequences
        sequence_length_distribution:
            14: 0.8 # 80% of all generated sequences for all repertoires will have length 14
            15: 0.2 # 20% of all generated sequences across all repertoires will have length 15
        labels:
            cmv: # label name
                True: 0.5 # 50% of the repertoires will have class True
                False: 0.5 # 50% of the repertoires will have class False
            coeliac: # next label with classes that will be assigned to repertoires independently of the previous label or any other parameter
                1: 0.3 # 30% of the generated repertoires will have class 1
                0: 0.7 # 70% of the generated repertoires will have class 0
        """
        RandomDatasetGenerator._check_rep_dataset_generation_params(repertoire_count, sequence_count_probabilities,
                                                                    sequence_length_probabilities,
                                                                    labels, path)

        alphabet = EnvironmentSettings.get_sequence_alphabet()
        PathBuilder.build(path)

        sequences = [["".join(random.choices(alphabet,
                                             k=random.choices(list(sequence_length_probabilities.keys()),
                                                              sequence_length_probabilities.values())[0]))
                      for seq_count in range(
                random.choices(list(sequence_count_probabilities.keys()), sequence_count_probabilities.values())[0])]
                     for rep in range(repertoire_count)]

        processed_labels, dataset_params = RandomDatasetGenerator._make_labels(labels, repertoire_count)

        dataset = RepertoireBuilder.build_dataset(sequences=sequences, path=path, labels=processed_labels, name=name)

        AIRRExporter.export(dataset, path)

        return dataset

    @staticmethod
    def _make_labels(labels: dict, element_count: int):
        if labels is not None:
            processed_labels = {
                label: random.choices(list(labels[label].keys()), labels[label].values(), k=element_count) for label
                in labels}
            dataset_params = {label: list(labels[label].keys()) for label in labels}
        else:
            processed_labels = None
            dataset_params = None

        return processed_labels, dataset_params

    @staticmethod
    def _check_receptor_dataset_generation_params(receptor_count: int, chain_1_length_probabilities: dict,
                                                  chain_2_length_probabilities: dict, labels: dict, path: Path):

        RandomDatasetGenerator._check_probabilities(chain_1_length_probabilities, int, 'chain_1_length_probabilities')
        RandomDatasetGenerator._check_probabilities(chain_2_length_probabilities, int, 'chain_2_length_probabilities')
        RandomDatasetGenerator._check_example_count(receptor_count, "receptor_count")
        RandomDatasetGenerator._check_labels(labels)
        RandomDatasetGenerator._check_path(path)

    @staticmethod
    def generate_receptor_dataset(receptor_count: int, chain_1_length_probabilities: dict,
                                  chain_2_length_probabilities: dict, labels: dict,
                                  path: Path, name="receptor_dataset"):
        """
        Creates receptor_count receptors where the length of sequences in each chain is sampled independently for each sequence from
        chain_n_length_probabilities distribution. The labels are also randomly assigned to receptors from the distribution given in
        labels. In this case, labels are multi-class, so each receptor will get one class from each label. This means that negative
        classes for the labels should be included as well in the specification. chain 1 and 2 in this case refer to alpha and beta
        chain of a T-cell receptor.

        An example of input parameters is given below:

        receptor_count: 100 # generate 100 TRABReceptors
        chain_1_length_probabilities:
            14: 0.8 # 80% of all generated sequences for all receptors (for chain 1) will have length 14
            15: 0.2 # 20% of all generated sequences across all receptors (for chain 1) will have length 15
        chain_2_length_probabilities:
            14: 0.8 # 80% of all generated sequences for all receptors (for chain 2) will have length 14
            15: 0.2 # 20% of all generated sequences across all receptors (for chain 2) will have length 15
        labels:
            epitope1: # label name
                True: 0.5 # 50% of the receptors will have class True
                False: 0.5 # 50% of the receptors will have class False
            epitope2: # next label with classes that will be assigned to receptors independently of the previous label or other parameters
                1: 0.3 # 30% of the generated receptors will have class 1
                0: 0.7 # 70% of the generated receptors will have class 0
        """
        RandomDatasetGenerator._check_receptor_dataset_generation_params(receptor_count, chain_1_length_probabilities,
                                                                         chain_2_length_probabilities, labels, path)

        alphabet = EnvironmentSettings.get_sequence_alphabet()
        PathBuilder.build(path)

        get_random_sequence = lambda proba, chain, id: ReceptorSequence(
            sequence_aa="".join(random.choices(alphabet, k=random.choices(list(proba.keys()), proba.values())[0])),
            duplicate_count=random.randint(1, 10), v_call=chain + "V1-1*01", j_call=chain + "J1-1*01", locus=chain, cell_id=str(id))

        receptors = []
        for i in range(receptor_count):
            receptor_id = uuid.uuid4().hex
            receptors.append(Receptor(chain_1=get_random_sequence(chain_1_length_probabilities, "TRA", receptor_id),
                                      chain_2=get_random_sequence(chain_2_length_probabilities, "TRB", receptor_id),
                                      chain_pair=ChainPair["TRA_TRB"],
                                      cell_id=str(receptor_id), receptor_id=receptor_id,
                                      metadata={
                                          **{label: random.choices(list(label_dict.keys()), label_dict.values(), k=1)[0]
                                             for label, label_dict in labels.items()}, **{"subject": f"subj_{i + 1}"}}))

        for receptor in receptors:
            receptor.chain_1.metadata = copy.deepcopy(receptor.metadata)
            receptor.chain_2.metadata = copy.deepcopy(receptor.metadata)

        processed_labels, dataset_params = RandomDatasetGenerator._make_labels(labels, receptor_count)
        dataset = ReceptorDataset.build_from_objects(receptors=receptors, path=path, name=name,
                                                     labels=dataset_params)

        AIRRExporter.export(dataset, path)

        return dataset

    @staticmethod
    def _check_sequence_dataset_generation_params(receptor_count: int, length_probabilities: dict, labels: dict,
                                                  path: Path):
        RandomDatasetGenerator._check_probabilities(length_probabilities, int, 'length_probabilities')
        RandomDatasetGenerator._check_example_count(receptor_count, "receptor_count")
        RandomDatasetGenerator._check_labels(labels)
        RandomDatasetGenerator._check_path(path)

    @staticmethod
    def generate_sequence_dataset(sequence_count: int, length_probabilities: dict, labels: dict, path: Path,
                                  region_type: str = RegionType.IMGT_CDR3.name, name="sequence_dataset"):
        """
        Creates sequence_count receptor sequences (single chain) where the length of sequences in each chain is sampled independently for each sequence from
        length_probabilities distribution. The labels are also randomly assigned to sequences from the distribution given in
        labels. In this case, labels are multi-class, so each sequences will get one class from each label. This means that negative
        classes for the labels should be included as well in the specification.

        An example of input parameters is given below:

        sequence_count: 100 # generate 100 TRB ReceptorSequences
        length_probabilities:
            14: 0.8 # 80% of all generated sequences for all receptors (for chain 1) will have length 14
            15: 0.2 # 20% of all generated sequences across all receptors (for chain 1) will have length 15
        labels:
            epitope1: # label name
                True: 0.5 # 50% of the receptors will have class True
                False: 0.5 # 50% of the receptors will have class False
            epitope2: # next label with classes that will be assigned to receptors independently of the previous label or other parameters
                1: 0.3 # 30% of the generated receptors will have class 1
                0: 0.7 # 70% of the generated receptors will have class 0
        """
        RandomDatasetGenerator._check_sequence_dataset_generation_params(sequence_count, length_probabilities, labels,
                                                                         path)

        alphabet = EnvironmentSettings.get_sequence_alphabet()
        PathBuilder.build(path)

        chain = "TRB"

        sequences = [ReceptorSequence(uuid.uuid4().hex, None,
                                      "".join(random.choices(alphabet,
                                                             k=random.choices(list(length_probabilities.keys()),
                                                                              length_probabilities.values())[0])),
                                      locus=chain, v_call=chain + "V1-1*01", j_call=chain + "J1-1*01",
                                      duplicate_count=random.randint(1, 10),
                                      metadata={
                                          **{label: random.choices(list(label_dict.keys()), label_dict.values(), k=1)[0]
                                             for label, label_dict in labels.items()}, **{"subject": f"subj_{i + 1}"}})
                     for i in range(sequence_count)]

        processed_labels, dataset_params = RandomDatasetGenerator._make_labels(labels, sequence_count)
        dataset = SequenceDataset.build_from_objects(sequences, path, name=name, labels=dataset_params,
                                                     region_type=RegionType[region_type.upper()])

        AIRRExporter.export(dataset, path)

        return dataset



import numpy as np
import torch
from torch import nn
from torch.nn.functional import relu

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType


class PyTorchReceptorCNN(nn.Module):
    """
    This class implements the ReceptorCNN using PyTorch. This is one specific implementation of the architecture proposed in
    :py:obj:`~immuneML.ml_methods.classifiers.ReceptorCNN.ReceptorCNN`.
    """

    def __init__(self, kernel_count: int, kernel_size, positional_channels: int, sequence_type: SequenceType, background_probabilities, chain_names):
        super(PyTorchReceptorCNN, self).__init__()
        self.background_probabilities = background_probabilities
        self.threshold = 0.1
        self.pseudocount = 0.05
        self.in_channels = len(EnvironmentSettings.get_sequence_alphabet(sequence_type)) + positional_channels
        self.positional_channels = positional_channels
        self.max_information_gain = self.get_max_information_gain()
        self.chain_names = chain_names

        self.conv_chain_1 = [f"chain_1_kernel_{size}" for size in kernel_size]
        self.conv_chain_2 = [f"chain_2_kernel_{size}" for size in kernel_size]

        for size in kernel_size:
            # chain 1
            setattr(self, f"chain_1_kernel_{size}", nn.Conv1d(in_channels=self.in_channels, out_channels=kernel_count, kernel_size=size,
                                                              bias=True))
            getattr(self, f"chain_1_kernel_{size}").weight.data. \
                normal_(0.0, np.sqrt(1 / np.prod(getattr(self, f"chain_1_kernel_{size}").weight.shape)))

            # chain 2
            setattr(self, f"chain_2_kernel_{size}", nn.Conv1d(in_channels=self.in_channels, out_channels=kernel_count, kernel_size=size,
                                                              bias=True))
            getattr(self, f"chain_2_kernel_{size}").weight.data. \
                normal_(0.0, np.sqrt(1 / np.prod(getattr(self, f"chain_2_kernel_{size}").weight.shape)))

        self.fully_connected = nn.Linear(in_features=kernel_count * len(kernel_size) * 2, out_features=1, bias=True)
        self.fully_connected.weight.data.normal_(0.0, np.sqrt(1 / np.prod(self.fully_connected.weight.shape)))

    def get_max_information_gain(self):
        """
        Information gain corresponds to Kullback-Leibler divergence between the observed probability p of an option (e.g. amino acid) ond null
        (or background) probability q of the option:

        .. math::

            KL(p||q) = \\sum_n p_n \\, log_2 \\, \\frac{p_n}{q_n}
                = \\sum_n p_n \\, log_2 \\, p_n - \\sum_n p_n \\, log_2 \\, q_n
                = \\sum_n p_n \\, log_2 \\, p_n - log_2 \\, q_n

            log_2 \\, q_n < 0 \\, \\, \\,  (1)

            \\sum_n p_n \\, log_2 \\, p_n < 0 \\, \\, \\,  (2)

            (1) \\wedge (2) \\Rightarrow max(KL(p||q)) = - log_2 \\, q_n

        Returns:

            max information gain given background probabilities

        """
        if all(self.background_probabilities[i] == self.background_probabilities[0] for i in range(len(self.background_probabilities))):
            return - np.log2(self.background_probabilities[0])
        else:
            raise NotImplementedError("ReceptorCNN: non-uniform background probabilities are currently not supported.")

    def forward(self, x):
        """
        Implements the forward pass through the network by applying kernels to the one-hot encoded receptors, followed by ReLU activation and max
        pooling. The obtained output is then concatenated to get the receptor representation. A fully-connected layer is then applied to the
        representation to predict the class assignment.

        Args:
            x: input data consisting of one-hot encoded immune receptors with optional positional information

        Returns:
            predictions of class assignment
        """

        # creates batch_size x kernel_count representation of chain 1 and chain 2 by applying kernels, followed by relu and global max pooling
        chain_1 = torch.cat([torch.max(relu(conv_kernel(x[:, 0])), dim=2)[0] for conv_kernel in
                             [getattr(self, name) for name in self.conv_chain_1]], dim=1)
        chain_2 = torch.cat([torch.max(relu(conv_kernel(x[:, 0])), dim=2)[0] for conv_kernel in
                             [getattr(self, name) for name in self.conv_chain_2]], dim=1)

        # creates a single representation of the receptor from chain representations
        receptor = torch.cat([chain_1, chain_2], dim=1).squeeze()

        # predict the class of the receptor through the final fully-connected layer based on the inferred representations
        predictions = self.fully_connected(receptor).squeeze()
        return predictions

    def rescale_weights_for_IGM(self):
        """Rescales the weights in the kernels to represent information gain matrices."""
        for name in self.conv_chain_1 + self.conv_chain_2:
            value = getattr(self, name)
            value.weight = self._rescale_chain(value.weight)

    def _rescale_chain(self, weight_parameter):
        # dimension for the content (without positional information):
        dim = self.in_channels - self.positional_channels

        # enforce non-negativity constraint
        weight_chain = relu(weight_parameter[:, :dim, :])

        # add pseudocount for positions where the total sum is below threshold
        weight_chain[torch.sum(weight_chain, dim=1, keepdim=True).expand_as(weight_chain) < self.threshold] += self.pseudocount

        # rescale chain weights to represent IGM
        weight_chain = weight_chain / torch.sum(weight_chain, dim=1, keepdim=True).expand_as(weight_chain) * self.max_information_gain

        # append positional kernel values to rescaled weights and convert back to parameter
        weight_parameter = nn.Parameter(torch.cat([weight_chain, weight_parameter[:, dim:, :]], dim=1))
        return weight_parameter



import logging
from pathlib import Path

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.hyperparameter_optimization import HPSetting
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.ml_metrics.MetricUtil import MetricUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class TrainingPerformance(MLReport):
    """
    A report that plots the evaluation metrics for the performance given machine learning model and training dataset.
    The available metrics are accuracy, balanced_accuracy, confusion_matrix, f1_micro, f1_macro, f1_weighted, precision,
    recall, auc and log_loss (see :py:obj:`immuneML.environment.Metric.Metric`).

    **Specification arguments:**

    - metrics (list): A list of metrics used to evaluate training performance. See :py:obj:`immuneML.environment.Metric.Metric` for available options.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_performance_report:
                    TrainingPerformance:
                        metrics:
                            - accuracy
                            - balanced_accuracy
                            - confusion_matrix
                            - f1_micro
                            - f1_macro
                            - f1_weighted
                            - precision
                            - recall
                            - auc
                            - log_loss

    """

    @classmethod
    def build_object(cls, **kwargs):
        location = "TrainingPerformance"        
        valid_metrics = [m.name for m in ClassificationMetric]

        name = kwargs["name"] if "name" in kwargs else None
        metrics = kwargs["metrics"] if "metrics" in kwargs else valid_metrics
        metrics = [m.upper() for m in metrics]
        
        ParameterValidator.assert_all_in_valid_list(metrics, valid_metrics, location, 'metrics')

        return TrainingPerformance(set(metrics), name=name)

    def __init__(self, metrics: set, train_dataset: Dataset = None, test_dataset: Dataset = None, method: MLMethod = None,
                 result_path: Path = None, name: str = None, hp_setting: HPSetting = None, label=None, number_of_processes: int = 1):
        super().__init__(train_dataset=train_dataset, test_dataset=test_dataset, method=method, result_path=result_path,
                         name=name, hp_setting=hp_setting, label=label, number_of_processes=number_of_processes)
        self.metrics_set = set(metrics)

    def _generate(self) -> ReportResult:
        
        X = self.train_dataset.encoded_data
        predicted_y = self.method.predict(X, self.label)[self.label.name]
        predicted_proba_y = self.method.predict_proba(X, self.label)[self.label.name][self.label.positive_class]
        true_y = self.train_dataset.encoded_data.labels[self.label.name]
        classes = self.method.get_classes()

        PathBuilder.build(self.result_path)

        scores = {}
        output = {
            'tables': [],
            'figures': []
        }

        for metric in self.metrics_set:
            _score = MetricUtil.score_for_metric(metric=ClassificationMetric.get_metric(metric),
                                                 predicted_y=predicted_y, predicted_proba_y=predicted_proba_y,
                                                 true_y=true_y, classes=classes)

            if metric == 'CONFUSION_MATRIX':
                self._generate_heatmap(classes, classes, _score, metric, output)
            else:
                scores[metric] = _score

        scores_df = pd.DataFrame.from_dict(scores, orient='index')
        scores_df.columns = [self.label.name]

        self._generate_barplot(scores_df, output)

        return ReportResult(self.name,
                            info="Plots the evaluation metrics for the performance given machine learning model and training dataset.",
                            output_tables=output['tables'],
                            output_figures=output['figures'])

    def _generate_barplot(self, df, output):
        import plotly.express as px

        path_csv = self.result_path / f"{self.name}.csv"
        path_html = self.result_path / f"{self.name}.html"

        df.to_csv(path_csv)

        figure = px.bar(df, x=df.index, y=self.label.name, labels={'index': "metrics"},
                        template='plotly_white', color_discrete_sequence=px.colors.diverging.Tealrose,
                        title=f"Evaluation metrics ({self.label})")

        figure.write_html(str(path_html))

        output['tables'].append(ReportOutput(path_csv, "training performance in csv"))
        output['figures'].append(ReportOutput(path_html, "training performance on selected metrics"))

        return

    def _generate_heatmap(self, x, y, z, metric, output, xlabel='Prediction', ylabel='Ground Truth', zlabel='Count'):
        path_csv = self.result_path / f"{self.name}_{metric.lower()}.csv"
        path_html = self.result_path / f"{self.name}_{metric.lower()}.html"
        
        z_flip = np.flipud(z)

        hovertext = []
        for yi, yy in enumerate(y):
            hovertext.append(list())
            for xi, xx in enumerate(x):
                hovertext[-1].append(f"{xlabel}: {xx}<br />{ylabel}: {yy}<br />{zlabel}: {z_flip[yi][xi]}")

        layout = go.Layout(
            title=f'Evaluation: {metric} ({self.label})',
            xaxis=dict(title=xlabel),
            yaxis=dict(title=ylabel)
        )
        trace = go.Heatmap(
            z=z_flip,
            x=x,
            y=y,
            hoverongaps = False,
            colorscale = 'burgyl',
            hoverinfo='text',
            text=hovertext
        )
        fig = go.Figure(data=[trace], layout=layout)
        fig.write_html(str(path_html))

        z_df = pd.DataFrame(z)
        z_df.columns = f'{xlabel} (' + pd.Index(map(str, x)) + ')'
        z_df.index = f'{ylabel} (' +  pd.Index(map(str, y)) + ')'
        z_df.to_csv(path_csv)   
        
        output['tables'].append(ReportOutput(path_csv, f"TrainingPerformance table ({metric.lower()})"))
        output['figures'].append(ReportOutput(path_html, f"TrainingPerformance html ({metric.lower()})"))
        
        return

    def check_prerequisites(self) -> bool:
        if not hasattr(self, "result_path") or self.result_path is None:
            logging.warning(f"{self.__class__.__name__} requires an output 'path' to be set. {self.__class__.__name__}"
                          f" report will not be created.")
            return False

        if self.train_dataset is None or self.train_dataset.encoded_data is None:
            logging.warning(
                f"{self.__class__.__name__}: train dataset is not encoded and can not be run."
                f"{self.__class__.__name__} report will not be created.")
            return False

        if self.method is None:
            logging.warning(
                f"{self.__class__.__name__}: method is not defined and can not be run."
                f"{self.__class__.__name__} report will not be created.")
            return False

        return True



from typing import List

from bionumpy.bnpdataclass import BNPDataClass

from immuneML.environment.SequenceType import SequenceType
from immuneML.simulation.SimConfigItem import SimConfigItem
from immuneML.simulation.implants.Signal import Signal
from immuneML.simulation.simulation_strategy.SimulationStrategy import SimulationStrategy
from immuneML.simulation.util.util import filter_out_illegal_sequences
from immuneML.util.Logger import print_log


class RejectionSamplingStrategy(SimulationStrategy):

    MAX_SIGNALS_PER_SEQUENCE = 2
    MAX_MOTIFS_PER_SEQUENCE = 1

    def process_sequences(self, sequences: BNPDataClass, seqs_per_signal_count: dict, use_p_gens: bool,
                          sequence_type: SequenceType, sim_item: SimConfigItem, all_signals: List[Signal],
                          remove_positives_first: bool, **kwargs) -> BNPDataClass:

        filtered_sequences = filter_out_illegal_sequences(sequences, sim_item, all_signals,
                                                          RejectionSamplingStrategy.MAX_SIGNALS_PER_SEQUENCE,
                                                          RejectionSamplingStrategy.MAX_MOTIFS_PER_SEQUENCE)

        removed_count = len(sequences) - len(filtered_sequences)
        if removed_count > 0:
            print_log(f"Removed {removed_count} out of {len(sequences)} during rejection sampling for having "
                      f"more than 1 signal.", True)

        return filtered_sequences


import copy
from pathlib import Path

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.Label import Label
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.hyperparameter_optimization.core.HPSelection import HPSelection
from immuneML.hyperparameter_optimization.core.HPUtil import HPUtil
from immuneML.hyperparameter_optimization.states.HPAssessmentState import HPAssessmentState
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.classifiers.SklearnMethod import SklearnMethod
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.reports.ReportUtil import ReportUtil
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.MLProcess import MLProcess


class HPAssessment:

    @staticmethod
    def run_assessment(state: TrainMLModelState) -> TrainMLModelState:

        state = HPAssessment._create_root_path(state)
        train_val_datasets, test_datasets = HPUtil.split_data(state.dataset, state.assessment, state.path,
                                                              state.label_configuration)
        n_splits = len(train_val_datasets)

        for index in range(n_splits):
            state = HPAssessment.run_assessment_split(state, train_val_datasets[index], test_datasets[index], index,
                                                      n_splits)

        return state

    @staticmethod
    def _create_root_path(state: TrainMLModelState) -> TrainMLModelState:
        name = state.name if state.name is not None else "result"
        state.path = state.path / name
        return state

    @staticmethod
    def run_assessment_split(state, train_val_dataset, test_dataset, split_index: int, n_splits):
        """run inner CV loop (selection) and retrain on the full train_val_dataset after optimal model is chosen"""

        print_log(f'Training ML model: running outer CV loop: started split {split_index + 1}/{n_splits}.\n',
                  include_datetime=True)

        current_path = HPAssessment.create_assessment_path(state, split_index)

        assessment_state = HPAssessmentState(split_index, train_val_dataset, test_dataset, current_path,
                                             state.label_configuration)
        state.assessment_states.append(assessment_state)

        state = HPSelection.run_selection(state, train_val_dataset, current_path, split_index)
        state = HPAssessment.run_assessment_split_per_label(state, split_index)

        assessment_state.train_val_data_reports = ReportUtil.run_data_reports(train_val_dataset,
                                                                              state.assessment.reports.data_split_reports.values(),
                                                                              current_path / "data_report_train",
                                                                              state.number_of_processes,
                                                                              state.context)
        assessment_state.test_data_reports = ReportUtil.run_data_reports(test_dataset,
                                                                         state.assessment.reports.data_split_reports.values(),
                                                                         current_path / "data_report_test",
                                                                         state.number_of_processes, state.context)

        print_log(f'Training ML model: running outer CV loop: finished split {split_index + 1}/{n_splits}.\n',
                  include_datetime=True)

        return state

    @staticmethod
    def run_assessment_split_per_label(state: TrainMLModelState, split_index: int):
        """iterate through labels and hp_settings and retrain all models"""
        n_labels = state.label_configuration.get_label_count()

        for idx, label in enumerate(state.label_configuration.get_label_objects()):

            print_log(f"Training ML model: running the inner loop of nested CV: "
                      f"retrain models for label {label.name} (label {idx + 1} / {n_labels}).\n", include_datetime=True)

            path = state.assessment_states[split_index].path

            for index, hp_setting in enumerate(state.hp_settings):

                if hp_setting != state.assessment_states[split_index].label_states[label.name].optimal_hp_setting:
                    setting_path = path / f"{label.name}_{hp_setting}/"
                else:
                    setting_path = path / f"{label.name}_{hp_setting}_optimal/"

                train_val_dataset = state.assessment_states[split_index].train_val_dataset
                test_dataset = state.assessment_states[split_index].test_dataset
                state = HPAssessment.reeval_on_assessment_split(state, train_val_dataset, test_dataset, hp_setting,
                                                                setting_path, label, split_index)

            print_log(f"Training ML model: running the inner loop of nested CV: completed retraining models "
                      f"for label {label.name} (label {idx + 1} / {n_labels}).\n", include_datetime=True)

        return state

    @staticmethod
    def reeval_on_assessment_split(state: TrainMLModelState, train_val_dataset: Dataset, test_dataset: Dataset,
                                   hp_setting: HPSetting, path: Path, label: Label, split_index: int) \
            -> TrainMLModelState:
        """retrain model for specific label, assessment split and hp_setting"""

        updated_hp_setting = HPAssessment.update_hp_setting_for_assessment(hp_setting, state, split_index, label.name)

        assessment_item = MLProcess(train_dataset=train_val_dataset, test_dataset=test_dataset, label=label,
                                    metrics=state.metrics, optimization_metric=state.optimization_metric, path=path,
                                    hp_setting=updated_hp_setting, report_context=state.context,
                                    ml_reports=state.assessment.reports.model_reports.values(),
                                    number_of_processes=state.number_of_processes,
                                    encoding_reports=state.assessment.reports.encoding_reports.values(),
                                    label_config=LabelConfiguration([label]),
                                    example_weighting=state.example_weighting,
                                    sequence_type=state.sequence_type, region_type=state.region_type).run(split_index)

        state.assessment_states[split_index].label_states[label.name].assessment_items[
            str(hp_setting)] = assessment_item

        return state

    @staticmethod
    def update_hp_setting_for_assessment(hp_setting: HPSetting, state: TrainMLModelState, split_index: int,
                                         label_name: str):

        if isinstance(hp_setting.ml_method, SklearnMethod) and hp_setting.ml_params['model_selection_cv']:
            updated_hp_setting = copy.deepcopy(hp_setting)
            updated_hp_setting.ml_params['model_selection_cv'] = False
            updated_hp_setting.ml_params['model_selection_n_folds'] = -1

            comp_func = ClassificationMetric.get_search_criterion(state.optimization_metric)
            hp_items = state.assessment_states[split_index].label_states[label_name].selection_state.hp_items[
                hp_setting.get_key()]

            if len(hp_items) > 1:
                optimal_params = {hp_item.performance[state.optimization_metric.name.lower()]:
                                      HPAssessment._get_only_hyperparams(hp_item.method.get_params())
                                  for hp_item in hp_items}
                updated_hp_setting.ml_params[updated_hp_setting.ml_method.__class__.__name__] = optimal_params[
                    comp_func(optimal_params.keys())]
                updated_hp_setting.ml_method = updated_hp_setting.ml_method.__class__(parameters=updated_hp_setting.ml_params[updated_hp_setting.ml_method.__class__.__name__])

            elif len(hp_items) == 1:
                updated_hp_setting.ml_params[updated_hp_setting.ml_method.__class__.__name__] = hp_items[
                    0].method.model.get_params()
                updated_hp_setting.ml_method = updated_hp_setting.ml_method.__class__(parameters=updated_hp_setting.ml_params[updated_hp_setting.ml_method.__class__.__name__])

            return updated_hp_setting

        else:
            return hp_setting

    @staticmethod
    def _get_only_hyperparams(params: dict):
        return copy.deepcopy({k: v for k, v in params.items() if k not in ['intercept', 'coefficients']})

    @staticmethod
    def create_assessment_path(state, split_index):
        current_path = state.path / f"split_{split_index + 1}"
        PathBuilder.build(current_path)
        return current_path



import logging

from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.kmer_frequency.sequence_encoding.SequenceEncodingStrategy import SequenceEncodingStrategy
from immuneML.environment.Constants import Constants
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.KmerHelper import KmerHelper


class IMGTGappedKmerEncoder(SequenceEncodingStrategy):

    @staticmethod
    def encode_sequence(sequence: ReceptorSequence, params: EncoderParams):
        """
        creates all overlapping gapped k-mers and IMGT position pairs from a sequence as features for use in KmerFrequencyEncoder.
        this gap length goes from min_gap to max_gap inclusive, and there is a k-mer of length k_left on the left
        side of the gap and a k-mer of length k_right on the right side of the gap.
        :param sequence: ReceptorSequence
        :param params: EncoderParams (within the "model", the following keys are used: "k_left", "k_right", "max_gap",
                        "min_gap")
        :return: SequenceEncodingResult
        """
        k_left = params.model.get('k_left')
        k_right = params.model.get('k_right', k_left)
        max_gap = params.model.get('max_gap')
        min_gap = params.model.get('min_gap', 0)
        sequence_type = params.model.get('sequence_type', EnvironmentSettings.sequence_type)
        length = len(getattr(sequence, params.get_seq_name_for_seq_object()))

        if length < k_left + k_right + max_gap:
            logging.warning('Sequence length is less than k_left + k_right + max_gap. Ignoring sequence')
            return None

        gapped_kmers = KmerHelper.create_IMGT_gapped_kmers_from_sequence(sequence, k_left=k_left, max_gap=max_gap,
                                                                         min_gap=min_gap, k_right=k_right,
                                                                         sequence_type=sequence_type,
                                                                         region_type=params.region_type)

        gapped_kmers = [Constants.FEATURE_DELIMITER.join([str(mer) for mer in kmer]) for kmer in gapped_kmers]

        return gapped_kmers

    @staticmethod
    def get_feature_names(params: EncoderParams):
        return ["sequence", "imgt_position"]



import abc
import logging
import os
import warnings
import inspect
from pathlib import Path

import dill
import numpy as np
import pkg_resources
import yaml
from sklearn.metrics import get_scorer_names
from sklearn.model_selection import RandomizedSearchCV
from sklearn.utils.validation import check_is_fitted

from immuneML.data_model.EncodedData import EncodedData
from immuneML.environment.Label import Label
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.util.Util import Util
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.util.FilenameHandler import FilenameHandler
from immuneML.util.PathBuilder import PathBuilder


class SklearnMethod(MLMethod):
    """
    Base class for ML methods imported from scikit-learn. The classes inheriting SklearnMethod acting as wrappers around imported
    ML methods from scikit-learn have to implement:
    - the __init__() method,
    - get_params(label) and
    - _get_ml_model()

    Other methods can also be overwritten if needed.
    The arguments and specification described bellow applied for all classes inheriting SklearnMethod.

    **Specification arguments:**

    - parameters: a dictionary of parameters that will be directly passed to scikit-learn's class upon calling __init__()
      method; for detailed list see scikit-learn's documentation of the specific class inheriting SklearnMethod

    - parameter_grid: a dictionary of parameters which all have to be valid arguments for scikit-learn's corresponding class' __init__() method
      (same as parameters), but unlike parameters argument can contain list of values instead of one value; if this is specified and
      "model_selection_cv" is True (in the specification) or just if fit_by_cross_validation() is called, a grid search will be performed over
      these parameters and the optimal model will be kept


    **YAML specification:**

        definitions:
            ml_methods:
                ml_methods:
                    log_reg:
                        LogisticRegression: # name of the class inheriting SklearnMethod
                            # sklearn parameters (same names as in original sklearn class)
                            max_iter: 1000 # specific parameter value
                            penalty: l1
                            # Additional parameter that determines whether to print convergence warnings
                            show_warnings: True
                        # if any of the parameters under LogisticRegression is a list and model_selection_cv is True,
                        # a grid search will be done over the given parameters, using the number of folds specified in model_selection_n_folds,
                        # and the optimal model will be selected
                        model_selection_cv: True
                        model_selection_n_folds: 5
                    svm_with_cv:
                        SVM: # name of another class inheriting SklearnMethod
                            # sklearn parameters (same names as in original sklearn class)
                            alpha: 10
                            # Additional parameter that determines whether to print convergence warnings
                            show_warnings: True
                        # no grid search will be done
                        model_selection_cv: False

    """

    FIT_CV = "fit_CV"
    FIT = "fit"

    def __init__(self, parameter_grid: dict = None, parameters: dict = None):
        super(SklearnMethod, self).__init__()
        self.model = None

        if parameter_grid is not None and "show_warnings" in parameter_grid:
            self.show_warnings = parameter_grid.pop("show_warnings")[0]
        elif parameters is not None and "show_warnings" in parameters:
            self.show_warnings = parameters.pop("show_warnings")
        else:
            self.show_warnings = True

        self._parameter_grid = parameter_grid
        self._parameters = parameters

    def _fit(self, encoded_data: EncodedData, cores_for_training: int = 2):
        mapped_y = Util.map_to_new_class_values(encoded_data.labels[self.label.name], self.class_mapping)

        self.model = self._fit_model(encoded_data.examples, mapped_y, encoded_data.example_weights, cores_for_training)

    def _predict(self, encoded_data: EncodedData):
        self.check_is_fitted(self.label.name)

        predictions = self.apply_with_weights(self.model.predict,
                                              encoded_data.example_weights,
                                              X=encoded_data.examples)

        return {self.label.name: Util.map_to_old_class_values(np.array(predictions), self.class_mapping)}

    def _predict_proba(self, encoded_data: EncodedData):
        if self.can_predict_proba():
            probabilities = self.apply_with_weights(self.model.predict_proba, encoded_data.example_weights, X=encoded_data.examples)
            class_names = Util.map_to_old_class_values(self.model.classes_, self.class_mapping)

            return {self.label.name: {class_name: probabilities[:, i] for i, class_name in enumerate(class_names)}}
        else:
            logging.warning(f"{self.__class__.__name__}: cannot predict probabilities.")
            return None

    def _fit_model(self, X, y, w=None, cores_for_training: int = 1):
        self.model = self._get_ml_model(cores_for_training, X)

        if w is not None and not self._check_method_supports_example_weight(self.model.fit) and not self._check_method_supports_example_weight(self.model.predict):
            logging.warning(f"{self.__class__.__name__}: cannot fit this classifier with example weights, fitting without example weights instead... Example weights will still be applied when computing evaluation metrics after fitting.")

        if not self.show_warnings:
            warnings.simplefilter("ignore")
            os.environ["PYTHONWARNINGS"] = "ignore"

        self.model = self._get_ml_model(cores_for_training, X)
        self.apply_with_weights(self.model.fit, w, X=X, y=y)

        if not self.show_warnings:
            del os.environ["PYTHONWARNINGS"]
            warnings.simplefilter("always")

        return self.model

    def apply_with_weights(self, method, weights, **kwargs):
        """
        Can be used to run self.model.fit, self.model.predict or self.model.predict_proba with sample weights if supported

        :param method: self.model.fit, self.model.predict or self.model.predict_proba
        :return: the result of the supplied method
        """
        if weights is not None and self._check_method_supports_example_weight(method):
            return method(**kwargs, sample_weight=weights)
        else:
            return method(**kwargs)

    def _check_method_supports_example_weight(self, method):
        return "sample_weight" in inspect.signature(method).parameters

    def can_predict_proba(self) -> bool:
        return False

    def check_is_fitted(self, label_name: str):
        if self.label.name == label_name or label_name is None:
            return check_is_fitted(self.model, ["estimators_", "coef_", "estimator", "_fit_X", "dual_coef_", "classes_"], all_or_any=any)

    def _fit_by_cross_validation(self, encoded_data: EncodedData, number_of_splits: int, cores_for_training: int):

        mapped_y = Util.map_to_new_class_values(encoded_data.labels[self.label.name], self.class_mapping)

        self.model = self._fit_model_by_cross_validation(X=encoded_data.examples, y=mapped_y, w=encoded_data.example_weights,
                                                         number_of_splits=number_of_splits, cores_for_training=cores_for_training)

    def _fit_model_by_cross_validation(self, X, y, w, number_of_splits: int, cores_for_training: int):

        model = self._get_ml_model()
        scoring = ClassificationMetric.get_sklearn_score_name(ClassificationMetric.get_metric(self.optimization_metric))

        if scoring not in get_scorer_names():
            scoring = "balanced_accuracy"
            logging.warning(
                f"{self.__class__.__name__}: specified optimization metric ({self.optimization_metric}) is not defined as a sklearn scoring function, using {scoring} instead... ")

        if not self.show_warnings:
            warnings.simplefilter("ignore")
            os.environ["PYTHONWARNINGS"] = "ignore"

        self.model = RandomizedSearchCV(model, param_distributions=self._parameter_grid, cv=number_of_splits, n_jobs=cores_for_training,
                                        scoring=scoring, refit=True)

        self.model = self.apply_with_weights(self.model.fit, w, X=X, y=y)

        if not self.show_warnings:
            del os.environ["PYTHONWARNINGS"]
            warnings.simplefilter("always")

        self.model = self.model.best_estimator_  # do not leave RandomSearchCV object to be in models, use the best estimator instead

        return self.model

    def store(self, path: Path):
        PathBuilder.build(path)
        file_path = path / f"{self._get_model_filename()}.pickle"
        with file_path.open("wb") as file:
            dill.dump(self.model, file)

        params_path = path / f"{self._get_model_filename()}.yaml"

        with params_path.open("w") as file:
            desc = {
                **(self.get_params()),
                "feature_names": self.get_feature_names(),
                "classes": self.model.classes_.tolist(),
                "class_mapping": self.class_mapping,
            }

            if self.label is not None:
                desc["label"] = self.label.get_desc_for_storage()

            yaml.dump(desc, file)

    def _get_model_filename(self):
        return FilenameHandler.get_filename(self.__class__.__name__, "")

    def load(self, path: Path):
        name = f"{self._get_model_filename()}.pickle"
        file_path = path / name
        if file_path.is_file():
            with file_path.open("rb") as file:
                self.model = dill.load(file)
        else:
            raise FileNotFoundError(f"{self.__class__.__name__} model could not be loaded from {file_path}"
                                    f". Check if the path to the {name} file is properly set.")

        params_path = path / f"{self._get_model_filename()}.yaml"

        if params_path.is_file():
            with params_path.open("r") as file:
                desc = yaml.safe_load(file)
                if "label" in desc:
                    setattr(self, "label", Label(**desc["label"]))
                for param in ["feature_names", "classes", "class_mapping"]:
                    if param in desc:
                        setattr(self, param, desc[param])

    @abc.abstractmethod
    def _get_ml_model(self, cores_for_training: int = 2, X=None):
        pass

    @abc.abstractmethod
    def get_params(self):
        """Returns the model parameters in a readable yaml-friendly way (consisting of lists, dictionaries and strings)."""
        pass

    def get_package_info(self) -> str:
        return Util.get_immuneML_version() + '; scikit-learn ' + pkg_resources.get_distribution('scikit-learn').version

    def get_compatible_encoders(self):
        from immuneML.encodings.evenness_profile.EvennessProfileEncoder import EvennessProfileEncoder
        from immuneML.encodings.kmer_frequency.KmerFrequencyEncoder import KmerFrequencyEncoder
        from immuneML.encodings.onehot.OneHotEncoder import OneHotEncoder
        from immuneML.encodings.word2vec.Word2VecEncoder import Word2VecEncoder
        from immuneML.encodings.reference_encoding.MatchedSequencesEncoder import MatchedSequencesEncoder
        from immuneML.encodings.reference_encoding.MatchedReceptorsEncoder import MatchedReceptorsEncoder
        from immuneML.encodings.reference_encoding.MatchedRegexEncoder import MatchedRegexEncoder
        from immuneML.encodings.motif_encoding.MotifEncoder import MotifEncoder

        return [KmerFrequencyEncoder, OneHotEncoder, Word2VecEncoder, EvennessProfileEncoder,
                MatchedSequencesEncoder, MatchedReceptorsEncoder, MatchedRegexEncoder, MotifEncoder]

    @staticmethod
    def get_usage_documentation(model_name):
        return f"""
    
    Scikit-learn models can be trained in two modes: 
    
    1. Creating a model using a given set of hyperparameters, and relying on the selection and assessment loop in the
    TrainMLModel instruction to select the optimal model. 
    
    2. Passing a range of different hyperparameters to {model_name}, and using a third layer of nested cross-validation 
    to find the optimal hyperparameters through grid search. In this case, only the {model_name} model with the optimal 
    hyperparameter settings is further used in the inner selection loop of the TrainMLModel instruction. 
    
    By default, mode 1 is used. In order to use mode 2, model_selection_cv and model_selection_n_folds must be set. 
    
    
    **Specification arguments:**

    - {model_name} (dict): Under this key, hyperparameters can be specified that will be passed to the scikit-learn class.
      Any scikit-learn hyperparameters can be specified here. In mode 1, a single value must be specified for each of the scikit-learn
      hyperparameters. In mode 2, it is possible to specify a range of different hyperparameters values in a list. It is also allowed
      to mix lists and single values in mode 2, in which case the grid search will only be done for the lists, while the
      single-value hyperparameters will be fixed. 
      In addition to the scikit-learn hyperparameters, parameter show_warnings (True/False) can be specified here. This determines
      whether scikit-learn warnings, such as convergence warnings, should be printed. By default show_warnings is True.
        
    - model_selection_cv (bool): If any of the hyperparameters under {model_name} is a list and model_selection_cv is True, 
      a grid search will be done over the given hyperparameters, using the number of folds specified in model_selection_n_folds.
      By default, model_selection_cv is False. 
        
    - model_selection_n_folds (int): The number of folds that should be used for the cross validation grid search if model_selection_cv is True.
        
    """



class NameBuilder:

    @staticmethod
    def build_name_from_dict(dictionary: dict, level=0):
        """ Creates a name from dictionary which includes all of its parameters and handles nested dictionaries up to depth of 10 inclusively

            Arguments:
                dictionary (dict): dictionary to create the name from
                level (int): controls recursion level, user should keep default

            Returns:
                name (str)
        """


        if level > 10:
            raise Exception("Too many nested dicts for serialization in NameBuilder!")

        name = ""

        keys = sorted(list(dictionary.keys()))
        for key in keys:

            if isinstance(dictionary[key], dict):
                value = "{" + NameBuilder.build_name_from_dict(dictionary[key], level=level + 1) + "}"
            else:
                value = str(dictionary[key])

            name = name + key + "_" + value + "__"

        name = name[0:(len(name) - 2)]
        name = name.replace(" ", "")

        return name



from pathlib import Path

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.preprocessing.filters.Filter import Filter


class ClonesPerRepertoireFilter(Filter):
    """
    Removes all repertoires from the RepertoireDataset, which contain fewer clonotypes than specified by the
    lower_limit, or more clonotypes than specified by the upper_limit.
    Note that this filter filters out repertoires, not individual sequences, and can thus only be applied to RepertoireDatasets.
    When no lower or upper limit is specified, or the value -1 is specified, the limit is ignored.

    Since the filter removes repertoires from the dataset (examples in machine learning setting), it cannot be used with :ref:`TrainMLModel`
    instruction. If you want to use this filter, see :ref:`DatasetExport` instruction with preprocessing.

    **Specification arguments:**

    - lower_limit (int): The minimal inclusive lower limit for the number of clonotypes allowed in a repertoire.

    - upper_limit (int): The maximal inclusive upper limit for the number of clonotypes allowed in a repertoire.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - my_filter:
                    ClonesPerRepertoireFilter:
                        lower_limit: 100
                        upper_limit: 100000

    """

    def __init__(self, result_path: Path = None, lower_limit: int = -1, upper_limit: int = -1):
        super().__init__(result_path)
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

    def keeps_example_count(self) -> bool:
        return False

    def process_dataset(self, dataset: RepertoireDataset, result_path: Path, number_of_processes=1):
        self.check_dataset_type(dataset, [RepertoireDataset], "ClonesPerRepertoireFilter")
        self.result_path = result_path if result_path is not None else self.result_path

        repertoires, indices = [], []

        for index, repertoire in enumerate(dataset.get_data()):
            if self.lower_limit != -1 and len(repertoire.data) < self.lower_limit:
                continue
            if self.upper_limit != -1 and len(repertoire.data) > self.upper_limit:
                continue
            repertoires.append(dataset.repertoires[index])
            indices.append(index)

        processed_dataset = RepertoireDataset.build_from_objects(repertoires=repertoires, path=result_path)

        self.check_dataset_not_empty(processed_dataset, "ClonesPerRepertoireFilter")

        return processed_dataset



import pandas as pd

from immuneML.data_model.datasets.ElementDataset import ReceptorDataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.util.EncoderHelper import EncoderHelper


class TCRdistEncoder(DatasetEncoder):
    """
    Encodes the given ReceptorDataset as a distance matrix between all receptors, where the distance is computed using TCRdist from the paper:
    Dash P, Fiore-Gartland AJ, Hertz T, et al. Quantifiable predictive features define epitope-specific T cell receptor repertoires.
    Nature. 2017; 547(7661):89-93. `doi:10.1038/nature22383 <https://www.nature.com/articles/nature22383>`_.

    For the implementation, `TCRdist3 <https://tcrdist3.readthedocs.io/en/latest/>`_ library was used (source code available
    `here <https://github.com/kmayerb/tcrdist3>`_).

    **Dataset type:**

    - ReceptorDatasets


    **Specification arguments:**

    - cores (int): number of processes to use for the computation


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_tcr_dist_enc:
                    TCRdist:
                        cores: 4

    """

    def __init__(self, cores: int, name: str = None):
        super().__init__(name=name)
        self.cores = cores
        self.distance_matrix = None
        self.context = None

    @staticmethod
    def build_object(dataset, **params):
        if isinstance(dataset, ReceptorDataset):
            return TCRdistEncoder(**params)
        else:
            raise ValueError("TCRdistEncoder is not defined for dataset types which are not ReceptorDataset.")

    def set_context(self, context: dict):
        self.context = context
        return self

    def encode(self, dataset, params: EncoderParams):
        train_receptor_ids = EncoderHelper.prepare_training_ids(dataset, params)
        if params.learn_model:
            self._build_tcr_dist_matrix(dataset, params.label_config.get_labels_by_name())

        distance_matrix = self.distance_matrix.loc[dataset.get_example_ids(), train_receptor_ids]
        labels = self._build_labels(dataset, params) if params.encode_labels else None

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(examples=distance_matrix, labels=labels, example_ids=distance_matrix.index.values,
                                                   encoding=TCRdistEncoder.__name__)

        return encoded_dataset

    def _build_tcr_dist_matrix(self, dataset: ReceptorDataset, label_names):
        from immuneML.util.TCRdistHelper import TCRdistHelper

        current_dataset = dataset if self.context is None or "dataset" not in self.context else self.context["dataset"]
        tcr_rep = TCRdistHelper.compute_tcr_dist(current_dataset, label_names, self.cores)
        self.distance_matrix = pd.DataFrame(tcr_rep.pw_alpha + tcr_rep.pw_beta, index=tcr_rep.clone_df.clone_id.values,
                                            columns=tcr_rep.clone_df.clone_id.values)

    def _build_labels(self, dataset: ReceptorDataset, params: EncoderParams) -> dict:
        labels = {label: [] for label in params.label_config.get_labels_by_name()}
        for receptor in dataset.get_data():
            for label_name in labels.keys():
                labels[label_name].append(receptor.metadata[label_name])
        return labels




import os
from pathlib import Path

import yaml

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.api.galaxy.GalaxyTool import GalaxyTool
from immuneML.api.galaxy.Util import Util
from immuneML.app.ImmuneMLApp import ImmuneMLApp
from immuneML.util.Logger import print_log
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.apply_gen_model.ApplyGenModelInstruction import ApplyGenModelInstruction


class ApplyGenModelTool(GalaxyTool):

    def __init__(self, specification_path: Path, result_path: Path, **kwargs):
        Util.check_parameters(specification_path, result_path, kwargs, ApplyGenModelTool.__name__)
        super().__init__(specification_path, result_path, **kwargs)

    def _run(self):
        PathBuilder.build(self.result_path)
        self._check_specs()
        state = ImmuneMLApp(self.yaml_path, self.result_path).run()[0]

        Util.export_galaxy_dataset(state.generated_dataset, self.result_path)

        print_log("Run the generative model, the resulting dataset is exported.")


    def _check_specs(self):
        with open(self.yaml_path, "r") as file:
            specs = yaml.safe_load(file)

        instruction_name = Util.check_instruction_type(specs, ApplyGenModelTool.__name__,
                                                       ApplyGenModelInstruction.__name__[:-11])

        ParameterValidator.assert_keys_present(list(specs['instructions'][instruction_name].keys()),
                                               ["ml_config_path", 'gen_examples_count'],
                                               ApplyGenModelTool.__name__, instruction_name)

        assert os.path.isfile(specs['instructions'][instruction_name]['ml_config_path']), \
            f"{ApplyGenModelTool.__name__}: file specified under 'ml_config_path' parameter " \
            f"({specs['instructions'][instruction_name]['ml_config_path']}) is not available. Please check if it was " \
            f"correctly uploaded or if the file name is correct."



from pathlib import Path
from typing import List

import pandas as pd
import plotly.graph_objects as go

from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.clustering_reports.ClusteringReport import ClusteringReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringState


class ExternalLabelClusterSummary(ClusteringReport):
    """
    This report summarizes the number of examples in a cluster with different values of external labels.
    For each external label, it creates:
    1. A contingency table showing the count of examples for each combination of cluster and label value
    2. A heatmap visualization of these counts

    **Specification arguments:**

    - external_labels (list): the list of metadata columns in the dataset that should be compared against cluster
      assignment

    """

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_keys(list(kwargs.keys()), ['external_labels', 'name'],
                                       ExternalLabelClusterSummary.__name__, ExternalLabelClusterSummary.__name__)
        ParameterValidator.assert_all_type_and_value(kwargs['external_labels'], str,
                                                     ExternalLabelClusterSummary.__name__, 'external_labels')
        return ExternalLabelClusterSummary(**kwargs)

    def __init__(self, external_labels: List[str], name: str = None, state: ClusteringState = None,
                 result_path: Path = None, number_of_processes: int = 1):
        super().__init__(name, result_path, number_of_processes, state)
        self.external_labels = external_labels
        self.desc = "External Label Cluster Summary"

    def _generate(self) -> ReportResult:
        self.result_path = PathBuilder.build(self.result_path / self.name)
        report_outputs = []

        # For each split in the clustering results
        for split_idx, clustering_results in enumerate(self.state.clustering_items):
            # Process discovery results
            if clustering_results.discovery:
                report_outputs.extend(self._process_analysis_results(
                    clustering_results.discovery, 
                    f"discovery_split_{split_idx + 1}"
                ))

            # Process method-based validation results if available
            if clustering_results.method_based_validation:
                report_outputs.extend(self._process_analysis_results(
                    clustering_results.method_based_validation,
                    f"method_based_validation_split_{split_idx + 1}"
                ))

            # Process result-based validation results if available
            if clustering_results.result_based_validation:
                report_outputs.extend(self._process_analysis_results(
                    clustering_results.result_based_validation,
                    f"result_based_validation_split_{split_idx + 1}"
                ))

        if not report_outputs:
            return ReportResult(
                name=f"{self.desc} ({self.name})",
                info="No results were generated. This could be because no external labels were found in the dataset "
                     "metadata."
            )

        return ReportResult(
            name=f"{self.desc} ({self.name})",
            info="Summary of cluster assignments versus external labels",
            output_tables=[output for output in report_outputs if 'table' in output.name],
            output_figures=[output for output in report_outputs if 'heatmap' in output.name]
        )

    def _process_analysis_results(self, analysis_results, analysis_name: str) -> List[ReportOutput]:
        outputs = []

        # For each clustering setting
        for setting_key, item_result in analysis_results.items.items():
            predictions = item_result.item.predictions
            dataset = item_result.item.dataset

            # For each external label
            labels = dataset.get_metadata(self.external_labels, return_df=True)
            for label in self.external_labels:
                label_values = labels[label]

                # Create contingency table
                contingency_df = pd.crosstab(
                    pd.Series(predictions, name='Cluster'),
                    pd.Series(label_values, name=label)
                )

                # Save contingency table
                table_path = self.result_path / f"{analysis_name}_{setting_key}_{label}_contingency.csv"
                contingency_df.to_csv(table_path)
                outputs.append(ReportOutput(
                    path=table_path,
                    name=f"Contingency table for {label} ({analysis_name.replace('_', ' ')}, {setting_key})"
                ))

                # Create heatmap
                fig = go.Figure(data=go.Heatmap(
                    z=contingency_df.values,
                    x=contingency_df.columns,
                    y=contingency_df.index,
                    colorscale='Tealrose',
                    text=contingency_df.values,
                    texttemplate='%{text}',
                    textfont={"size": 10},
                    hoverongaps=False
                ))

                fig.update_layout(
                    xaxis_title=label,
                    yaxis_title='cluster'
                )

                # Save heatmap
                heatmap_path = self.result_path / f"{analysis_name}_{setting_key}_{label}_heatmap.html"
                fig.write_html(str(heatmap_path))
                outputs.append(ReportOutput(
                    path=heatmap_path,
                    name=f"Distribution heatmap for {label} ({analysis_name.replace('_', ' ')}, {setting_key})"
                ))

        return outputs

    def check_prerequisites(self):
        if not self.state:
            return False

        if not self.state.clustering_items:
            return False

        if not self.external_labels:
            return False

        return True



import logging
import shutil
from pathlib import Path

import yaml

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.api.galaxy.GalaxyTool import GalaxyTool
from immuneML.api.galaxy.Util import Util
from immuneML.app.ImmuneMLApp import ImmuneMLApp
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.train_gen_model.TrainGenModelInstruction import TrainGenModelInstruction


class GalaxyTrainGenModel(GalaxyTool):

    def __init__(self, specification_path: Path, result_path: Path, **kwargs):
        Util.check_parameters(specification_path, result_path, kwargs, GalaxyTrainGenModel.__name__)
        super().__init__(specification_path, result_path, **kwargs)
        self.instruction_name = None

    def _run(self):
        PathBuilder.build(self.result_path)
        self._prepare_specs()
        app = ImmuneMLApp(self.yaml_path, self.result_path)
        state = app.run()[0]

        dataset = state.generated_dataset if state.combined_dataset is None else state.combined_dataset
        Util.export_galaxy_dataset(dataset, self.result_path)

        model_locations = list(self.result_path.glob(f"{self.instruction_name}/trained_model/*.zip"))
        model_export_path = PathBuilder.build(self.result_path / 'exported_models/')

        for model_location in model_locations:
            shutil.copyfile(model_location, model_export_path / model_location.name)

        logging.info(f"{GalaxyTrainGenModel.__name__}: immuneML has finished and the trained models and dataset were exported.")


    def _prepare_specs(self):
        with self.yaml_path.open("r") as file:
            specs = yaml.safe_load(file)

        location = GalaxyTrainGenModel.__name__

        ParameterValidator.assert_keys_present(specs.keys(), ["definitions", "instructions"], location, "YAML specification")
        ParameterValidator.assert_all_in_valid_list(specs.keys(), ["definitions", "instructions", "output"], location,
                                                    "YAML specification")

        ParameterValidator.assert_type_and_value(specs["instructions"], dict, location, "instructions")

        assert len(list(specs["instructions"].keys())) == 1, f"{location}: one instruction has to be specified under " \
                                                             f"`instructions`, got the following instead: {list(specs['instructions'].keys())}."

        self.instruction_name = list(specs["instructions"].keys())[0]

        ParameterValidator.assert_type_and_value(specs['instructions'][self.instruction_name], dict, location,
                                                 self.instruction_name)
        ParameterValidator.assert_keys_present(specs['instructions'][self.instruction_name].keys(), ['type'], location,
                                               self.instruction_name)

        assert specs['instructions'][self.instruction_name]['type'] == TrainGenModelInstruction.__name__[:-11], \
            f"{GalaxyTrainGenModel.__name__}: instruction `type` under {self.instruction_name} has to be {TrainGenModelInstruction.__name__[:-11]} " \
            f"for this tool."

        Util.check_paths(specs, location)
        Util.update_result_paths(specs, self.result_path, self.yaml_path)



import logging
import os
import subprocess
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.cluster.hierarchy import dendrogram
from sklearn.cluster import AgglomerativeClustering

from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.PathBuilder import PathBuilder


class CompAIRRClusteringReport(DataReport):
    """
    A report that uses CompAIRR to compute repertoire distances based on sequence overlap and performs hierarchical
    clustering on the resulting distance matrix. The clustering results are visualized using a dendrogram,
    colored by a specified label.

    The report assumes that CompAIRR (https://github.com/uio-bmi/compairr) has been installed.

    Arguments:

        label (str): The label by which to color the dendrogram leaves. Must be one of the metadata fields in the dataset.

        compairr_path (str): Path to the CompAIRR executable.

        max_distance (int): Max Hamming distance allowed between sequences (default: 1)

        indels (bool): Whether to allow insertions/deletions when matching sequences (default: False)

        ignore_counts (bool): Whether to ignore sequence counts when computing overlap (default: False)

        ignore_genes (bool): Whether to ignore V/J gene assignments when matching sequences (default: False)

        threads (int): Number of threads to use for CompAIRR computation (default: 4)

        linkage_method (str): The linkage method to use for hierarchical clustering (default: 'single')

        is_cdr3 (bool): Whether the sequences represent CDR3s (default: True)

        clustering_threshold (float): The threshold to use for cutting the dendrogram into clusters (default: 0.5)

    """

    @classmethod
    def build_object(cls, **kwargs):
        return CompAIRRClusteringReport(**kwargs)

    def __init__(self, dataset: RepertoireDataset = None, result_path: Path = None, label: str = None,
                 compairr_path: str = None,
                 max_distance: int = 1, indels: bool = False, ignore_counts: bool = False, ignore_genes: bool = False,
                 threads: int = 4, linkage_method: str = 'single', is_cdr3: bool = True,
                 clustering_threshold: float = 0.5, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, name=name)
        self.label = label
        self.linkage_method = linkage_method
        self.compairr_path = compairr_path
        self.max_distance = max_distance
        self.indels = indels
        self.ignore_counts = ignore_counts
        self.ignore_genes = ignore_genes
        self.threads = threads
        self.is_cdr3 = is_cdr3
        self.clustering_threshold = clustering_threshold

    def check_prerequisites(self) -> bool:
        if not self.compairr_path:
            logging.warning("CompAIRR path not provided. CompAIRR must be installed and available in the system PATH.")
            return False
        if not isinstance(self.dataset, RepertoireDataset):
            logging.warning("CompAIRRClusteringReport requires a RepertoireDataset as input.")
            return False
        return True

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        self.check_label()

        similarity_matrix = self.compare_repertoires()
        distance_matrix = self.compute_distance_matrix(similarity_matrix)

        model = AgglomerativeClustering(distance_threshold=0, n_clusters=None, linkage=self.linkage_method,
                                        metric='precomputed')
        predictions = model.fit_predict(distance_matrix.values)

        # Save outputs
        distance_output = self._store_matrix(distance_matrix, 'distance_matrix')
        similarity_output = self._store_matrix(similarity_matrix, 'similarity_matrix')
        dendrogram_output = self._create_dendrogram(model, self.dataset.get_metadata(['subject_id'])['subject_id'])
        cluster_output = self._store_cluster_assignments(self.dataset.get_example_ids(), predictions)

        return ReportResult(
            name=self.name,
            info="Hierarchical clustering of repertoires based on CompAIRR sequence overlap",
            output_figures=[dendrogram_output],
            output_tables=[distance_output, similarity_output, cluster_output]
        )

    def compute_distance_matrix(self, similarity_matrix) -> pd.DataFrame:
        return pd.DataFrame(data=1 - similarity_matrix.values / similarity_matrix.values.max(),
                            columns=similarity_matrix.columns, index=similarity_matrix.index)

    def compare_repertoires(self) -> pd.DataFrame:
        # Create similarity matrix
        n_repertoires = len(self.dataset.repertoires)
        similarity_matrix = self._init_similarity_matrix(n_repertoires)

        # Compare repertoires pairwise
        for i in range(n_repertoires):
            rep1 = self.dataset.repertoires[i]
            similarity_matrix.loc[rep1.identifier, rep1.identifier] = rep1.get_element_count()
            for j in range(i + 1, n_repertoires):
                rep2 = self.dataset.repertoires[j]
                similarity_matrix = self._run_compairr(rep1, rep2, similarity_matrix)

        return similarity_matrix

    def check_label(self):
        if self.label not in self.dataset.get_label_names():
            raise ValueError(
                f"Label '{self.label}' not found in dataset metadata. Available labels: {self.dataset.get_label_names()}")

    def _init_similarity_matrix(self, n_repertoires: int) -> pd.DataFrame:
        return pd.DataFrame(
            np.zeros((n_repertoires, n_repertoires)),
            index=[rep.identifier for rep in self.dataset.repertoires],
            columns=[rep.identifier for rep in self.dataset.repertoires]
        )

    def _run_compairr(self, rep1: Repertoire, rep2: Repertoire, similarity_matrix: pd.DataFrame):
        output_file = self.result_path / f"overlap_{rep1.identifier}_{rep2.identifier}.tsv"

        cmd_args = [str(self.compairr_path), "-m", str(rep1.data_filename), str(rep2.data_filename),
                    "-d", str(self.max_distance), "-o", str(output_file)]
        if self.indels:
            cmd_args.append("-i")
        if self.ignore_counts:
            cmd_args.append("-f")
        if self.ignore_genes:
            cmd_args.append("-g")
        if self.is_cdr3:
            cmd_args.append("--cdr3")

        subprocess.run(cmd_args, capture_output=True, text=True)

        # Calculate similarity as sum of matches
        if output_file.is_file():
            similarity = pd.read_csv(output_file, sep="\t").sum().sum()
            similarity_matrix.loc[rep1.identifier, rep2.identifier] = similarity
            similarity_matrix.loc[rep2.identifier, rep1.identifier] = similarity

            os.remove(str(output_file))

        return similarity_matrix

    def _store_matrix(self, matrix: pd.DataFrame, name: str) -> ReportOutput:
        output_path = self.result_path / f"{name}.tsv"
        matrix.to_csv(output_path, sep="\t")
        return ReportOutput(
            path=output_path,
            name=name,
        )

    def _store_cluster_assignments(self, repertoire_ids, clusters) -> ReportOutput:
        output_path = self.result_path / "cluster_assignments.tsv"
        cluster_df = pd.DataFrame({
            'repertoire_id': repertoire_ids,
            'cluster': clusters
        })
        cluster_df.to_csv(output_path, sep="\t", index=False)
        return ReportOutput(
            path=output_path,
            name="Cluster Assignments",
        )

    def _create_dendrogram(self, model, repertoire_ids: list) -> ReportOutput:
        from matplotlib import pyplot as plt
        fig = plt.figure(figsize=(len(repertoire_ids) * 0.3 + 1, 10))

        # create the counts of samples under each node
        counts = np.zeros(model.children_.shape[0])
        n_samples = len(model.labels_)
        for i, merge in enumerate(model.children_):
            current_count = 0
            for child_idx in merge:
                if child_idx < n_samples:
                    current_count += 1  # leaf node
                else:
                    current_count += counts[child_idx - n_samples]
            counts[i] = current_count

        linkage_matrix = np.column_stack([model.children_, model.distances_, counts]).astype(float)

        labels = self.dataset.get_metadata([self.label])[self.label]
        import plotly.express as px
        unique_label_values = list(self.dataset.labels[self.label])
        colors = {rep_id: tuple([round(float(el) / 255, 3) for el in
                                 px.colors.qualitative.Prism[unique_label_values.index(labels[ind])][4:-1].split(', ')])
                  for ind, rep_id in enumerate(repertoire_ids)}

        ddata = dendrogram(linkage_matrix, labels=np.array(repertoire_ids))

        for leaf, leaf_color in zip(plt.gca().get_xticklabels(), ddata["leaves_color_list"]):
            leaf.set_color(colors[leaf.get_text()])

        plt.xticks(rotation=90)

        output_path = self.result_path / "hierarchical_clustering_of_repertoires.png"
        plt.savefig(str(output_path), dpi=300)

        return ReportOutput(path=output_path, name="hierarchical_clustering_of_repertoires")



import os
import sys

sys.path.insert(0, os.path.abspath(os.getcwd()))

from immuneML.dsl.InstructionParser import InstructionParser
from immuneML.dsl.OutputParser import OutputParser
from immuneML.dsl.definition_parsers.DefinitionParser import DefinitionParser
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.PathBuilder import PathBuilder


def generate_docs(docs_path: str):
    docs_path = PathBuilder.remove_old_and_build(docs_path)

    DefinitionParser.generate_docs(docs_path)
    InstructionParser.generate_docs(docs_path)
    OutputParser.generate_docs(docs_path)
    print(f"Specification documentation is generated at {docs_path}.")


if __name__ == "__main__":
    if len(sys.argv) == 2:
        path = sys.argv[1]
    else:
        path = EnvironmentSettings.specs_docs_path

    generate_docs(path)



import logging
from pathlib import Path
from typing import List, Tuple

import pandas as pd
import plotly.express as px

from immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder import CompAIRRSequenceAbundanceEncoder
from immuneML.encodings.abundance_encoding.KmerAbundanceEncoder import KmerAbundanceEncoder
from immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder import SequenceAbundanceEncoder
from immuneML.environment.Label import Label
from immuneML.hyperparameter_optimization.states.HPItem import HPItem
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.train_ml_model_reports.TrainMLModelReport import TrainMLModelReport
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.SequenceAnalysisHelper import SequenceAnalysisHelper


class DiseaseAssociatedSequenceCVOverlap(TrainMLModelReport):
    """
    DiseaseAssociatedSequenceCVOverlap report makes one heatmap per label showing the overlap of disease-associated sequences (or k-mers)
    produced by the :py:obj:`~immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder.SequenceAbundanceEncoder`,
    :py:obj:`~immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder.CompAIRRSequenceAbundanceEncoder` or
    :py:obj:`~immuneML.encodings.abundance_encoding.KmerAbundanceEncoder.KmerAbundanceEncoder`
    between folds of cross-validation (either inner or outer loop of the nested CV). The overlap is computed by the following equation:

    .. math::

        overlap(X,Y) = \\frac{|X \\cap Y|}{min(|X|, |Y|)} x 100

    For details, see Greiff V, Menzel U, Miho E, et al. Systems Analysis Reveals High Genetic and Antigen-Driven Predetermination of Antibody
    Repertoires throughout B Cell Development. Cell Reports. 2017;19(7):1467-1478. doi:10.1016/j.celrep.2017.04.054.


    **Specification arguments:**

    - compare_in_selection (bool): whether to compute the overlap over the inner loop of the nested CV - the sequence overlap is shown across CV
      folds for the model chosen as optimal within that selection

    - compare_in_assessment (bool): whether to compute the overlap over the optimal models in the outer loop of the nested CV


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_overlap_report: DiseaseAssociatedSequenceCVOverlap # report has no parameters

    """
    COMPATIBLE_ENCODERS = (SequenceAbundanceEncoder, CompAIRRSequenceAbundanceEncoder, KmerAbundanceEncoder)

    @classmethod
    def build_object(cls, **kwargs):
        return DiseaseAssociatedSequenceCVOverlap(**kwargs)

    def __init__(self, state: TrainMLModelState = None, result_path: Path = None, name: str = None, compare_in_selection: bool = False,
                 compare_in_assessment: bool = False, label: Label = None, number_of_processes: int = 1):
        super().__init__(name=name, state=state, label=label, result_path=result_path, number_of_processes=number_of_processes)
        self.compare_in_selection = compare_in_selection
        self.compare_in_assessment = compare_in_assessment

    def _generate(self) -> ReportResult:

        assert self.state, (f"{self.__class__.__name__}: this report can only be used in combination with TrainMLModel "
                            f"instruction. The instruction output is not available here so the report will be skipped.")

        PathBuilder.build(self.result_path)

        tables, figures = [], []
        for label_name in self.state.label_configuration.get_labels_by_name():
            if self.compare_in_assessment:
                table, figure = self._generate_for_assessment(label_name)
                tables.append(table)
                figures.append(figure)
            if self.compare_in_selection:
                tmp_tables, tmp_figures = self._generate_for_selection(label_name)
                tables += tmp_tables
                figures += tmp_figures

        return ReportResult(self.name,
                            info="One heatmap per label showing the overlap of disease-associated sequences produced by the SequenceAbundance encoder between folds of cross-validation (either inner or outer loop of the nested CV)",
                            output_figures=[fig for fig in figures if fig is not None],
                            output_tables=[tab for tab in tables if tab is not None])

    def _generate_for_assessment(self, label_name: str):
        hp_items = [st.label_states[label_name].optimal_assessment_item for st in self.state.assessment_states
                    if st.label_states[label_name].optimal_assessment_item.encoder.__class__ in DiseaseAssociatedSequenceCVOverlap.COMPATIBLE_ENCODERS]
        table, figure = self._compute_overlap(hp_items, f'sequence_overlap_{label_name}_assessment')
        return table, figure

    def _generate_for_selection(self, label_name: str):
        tables, figures = [], []
        for assessment_index, assessment_state in enumerate(self.state.assessment_states):
            selection_state = assessment_state.label_states[label_name].selection_state
            if selection_state.optimal_hp_setting.encoder.__class__ in DiseaseAssociatedSequenceCVOverlap.COMPATIBLE_ENCODERS:
                hp_items = selection_state.hp_items[selection_state.optimal_hp_setting.get_key()]
                table, figure = self._compute_overlap(hp_items, f'sequence_overlap_{label_name}_selection_{assessment_index + 1}_split')
                tables.append(table)
                figures.append(figure)
        return tables, figures

    def _compute_overlap(self, hp_items: List[HPItem], filename: str) -> Tuple[ReportOutput, ReportOutput]:
        overlap_matrix = SequenceAnalysisHelper.compute_overlap_matrix(hp_items)

        if overlap_matrix is None:
            logging.warning(f'{DiseaseAssociatedSequenceCVOverlap.__name__}: overlap matrix is None, some of the relevant sequence sets were empty, '
                            f'no report will be made.')
            return None, None

        row_col_names = [f"{item.hp_setting}_split_{item.split_index+1}" for item in hp_items]
        table_output = self._export_matrix(overlap_matrix, filename, row_col_names)
        figure_output = self._make_figure(overlap_matrix, filename, row_col_names)
        return table_output, figure_output

    def _export_matrix(self, overlap_matrix, filename, row_col_names) -> ReportOutput:
        data_path = self.result_path / f"{filename}.csv"
        pd.DataFrame(overlap_matrix, columns=row_col_names, index=row_col_names).to_csv(data_path)
        return ReportOutput(data_path, " ".join(filename.split('_') + ['data']))

    def _make_figure(self, overlap_matrix, filename, row_col_names) -> ReportOutput:
        figure = px.imshow(overlap_matrix, x=row_col_names, y=row_col_names, zmin=0, zmax=100, color_continuous_scale=px.colors.sequential.Teal,
                           template='plotly_white')
        figure.update_traces(hovertemplate="Overlap of disease-associated<br>sequences between<br>%{x} and %{y}:<br>%{z}%<extra></extra>")
        figure_path = self.result_path / f"{filename}.html"
        figure.write_html(str(figure_path))
        return ReportOutput(figure_path, " ".join(filename.split('_')))



import logging

import numpy as np
import plotly.graph_objs as go
from sklearn.metrics import roc_curve, auc

from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.PathBuilder import PathBuilder
from immuneML.ml_methods.util.Util import Util


class ROCCurve(MLReport):
    """
    A report that plots the ROC curve for a binary classifier.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_roc_report: ROCCurve

    """

    @classmethod
    def build_object(cls, **kwargs):
        name = kwargs["name"] if "name" in kwargs else "ROC_curve"
        return ROCCurve(name=name)

    def _generate(self) -> ReportResult:
        x = self.test_dataset.encoded_data

        if self.method.can_predict_proba():
            predicted_y = self.method.predict_proba(x, self.label)[self.label.name][self.label.positive_class]
        else:
            predicted_y = self.method.predict(x, self.label)[self.label.name]
            predicted_y = Util.map_to_new_class_values(predicted_y, self.method.class_mapping)

        true_y = x.labels[self.label.name]
        true_y = Util.map_to_new_class_values(true_y, self.method.class_mapping)

        fpr, tpr, _ = roc_curve(true_y, predicted_y)
        roc_auc = auc(fpr, tpr)


        trace1 = go.Scatter(x=fpr, y=tpr,
                            mode='lines',
                            line=dict(color='darkorange', width=2),
                            name=f"ROC curve (area = {roc_auc})")
        trace2 = go.Scatter(x=[0, 1], y=[0, 1],
                            mode='lines',
                            line=dict(color='navy', width=2, dash='dash'),
                            showlegend=False)
        layout = go.Layout(title='Receiver operating characteristic example',
                           xaxis=dict(title='False Positive Rate'),
                           yaxis=dict(title='True Positive Rate'),
                           template="plotly_white")

        fig = go.Figure(data=[trace1, trace2], layout=layout)

        PathBuilder.build(self.result_path)
        path_htm = self.result_path / f"{self.name}.html"
        path_csv = self.result_path / f"{self.name}.csv"
        csv_result = np.concatenate((fpr.reshape(1, -1), tpr.reshape(1, -1)))
        fig.write_html(str(path_htm))
        np.savetxt(str(path_csv), csv_result, header="fpr,tpr")
        return ReportResult(self.name,
                            info="A report that plots the ROC curve for a binary classifier.",
                            output_figures=[ReportOutput(path_htm)],
                            output_tables=[ReportOutput(path_csv)])

    def check_prerequisites(self):
        if not hasattr(self, "result_path") or self.result_path is None:
            logging.warning(f"{self.__class__.__name__} requires an output"
                          f" 'path' to be set. {self.__class__.__name__}"
                          f" report will not be created.")
            return False

        if self.test_dataset.encoded_data is None:
            logging.warning(
                f"{self.__class__.__name__}: test dataset is"
                f" not encoded and can not be run."
                f"{self.__class__.__name__} report will not be created.")
            return False

        if self.method is None:
            logging.warning(
                f"{self.__class__.__name__}: method is"
                f" not defined and can not be run."
                f"{self.__class__.__name__} report will not be created.")
        return True



from pathlib import Path

import pandas as pd

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.PathBuilder import PathBuilder


class DeepRCEncoder(DatasetEncoder):
    """
    DeepRCEncoder should be used in combination with the DeepRC ML method (:ref:`DeepRC`).
    This encoder writes the data in a RepertoireDataset to .tsv files.
    For each repertoire, one .tsv file is created containing the amino acid sequences and the counts.
    Additionally, one metadata .tsv file is created, which describes the subset of repertoires that is encoded by
    a given instance of the DeepRCEncoder.

    Note: sequences where count is None, the count value will be set to 1

    **Dataset type:**

    - RepertoireDatasets


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_deeprc_encoder: DeepRC

    """
    ID_COLUMN = "ID"
    SEQUENCE_COLUMN = "amino_acid"
    COUNTS_COLUMN = "templates"
    SEP = "\t"
    EXTENSION = "tsv"
    METADATA_EXTENSION = "csv"
    METADATA_SEP = ","

    def __init__(self, context: dict = None, name: str = None):
        super().__init__(name=name)
        self.context = context
        self.max_sequence_length = 0

    def set_context(self, context: dict):
        self.context = context
        return self

    @staticmethod
    def _prepare_parameters(context: dict = None, name: str = None):
        return {"context": context,
                "name": name}

    @staticmethod
    def build_object(dataset, **params):
        if isinstance(dataset, RepertoireDataset):
            prepared_params = DeepRCEncoder._prepare_parameters(**params)
            return DeepRCEncoder(**prepared_params)
        else:
            raise ValueError("DeepRCEncoder is not defined for dataset types which are not RepertoireDataset.")

    def export_repertoire_tsv_files(self, output_folder: Path, params: EncoderParams):
        repertoires = self.context["dataset"].repertoires
        sequence_column = params.region_type.value + ("_aa" if params.sequence_type == SequenceType.AMINO_ACID else "")

        for repertoire in repertoires:
            filepath = output_folder / f"{repertoire.identifier}.{DeepRCEncoder.EXTENSION}"

            if not filepath.is_file():
                df = (repertoire.data.topandas()[[sequence_column, 'duplicate_count']]
                      .rename(columns={sequence_column: DeepRCEncoder.SEQUENCE_COLUMN,
                                       'duplicate_count': DeepRCEncoder.COUNTS_COLUMN}))
                df[DeepRCEncoder.COUNTS_COLUMN].fillna(1, inplace=True)

                df.to_csv(path_or_buf=filepath, sep=DeepRCEncoder.SEP, index=False)

                max_sequence_length = max(df[DeepRCEncoder.SEQUENCE_COLUMN].str.len())
                self.max_sequence_length = max(self.max_sequence_length, max_sequence_length)

    def export_metadata_file(self, dataset, labels, output_folder):
        metadata_filepath = output_folder / f"{dataset.identifier}_metadata.{DeepRCEncoder.METADATA_EXTENSION}"
        metadata = dataset.get_metadata(labels, return_df=True)
        metadata[DeepRCEncoder.ID_COLUMN] = dataset.get_repertoire_ids()

        metadata.to_csv(path_or_buf=metadata_filepath, sep=DeepRCEncoder.METADATA_SEP, index=False)

        return metadata_filepath

    def encode(self, dataset, params: EncoderParams) -> RepertoireDataset:
        result_path = params.result_path / "encoding"
        PathBuilder.build(result_path)

        self.export_repertoire_tsv_files(result_path, params)

        metadata_filepath = self.export_metadata_file(dataset, params.label_config.get_labels_by_name(), result_path)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(examples=None,
                                                   labels=EncoderHelper.encode_dataset_labels(dataset,
                                                                                              params.label_config,
                                                                                              params.encode_labels),
                                                   example_ids=dataset.get_repertoire_ids(),
                                                   encoding=DeepRCEncoder.__name__,
                                                   info={"metadata_filepath": metadata_filepath,
                                                         "max_sequence_length": self.max_sequence_length})

        return encoded_dataset



import numpy as np

from immuneML.data_model.datasets.ElementDataset import ReceptorDataset, SequenceDataset
from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.workflows.steps.data_splitter.DataSplitterParams import DataSplitterParams
from immuneML.workflows.steps.data_splitter.Util import Util


class LeaveOneOutSplitter:

    @staticmethod
    def split_dataset(input_params: DataSplitterParams):
        if isinstance(input_params.dataset, ReceptorDataset) or isinstance(input_params.dataset, SequenceDataset):
            return LeaveOneOutSplitter._split_receptor_dataset(input_params)
        else:
            raise NotImplementedError("LeaveOneOutSplitter: leave-one-out stratification is currently implemented only for receptor dataset, "
                                      f"got {type(input_params.dataset).__name__} instead.")

    @staticmethod
    def _split_receptor_dataset(input_params: DataSplitterParams):
        dataset = input_params.dataset
        param, min_count = input_params.split_config.leave_one_out_config.parameter, input_params.split_config.leave_one_out_config.min_count

        unique_values = LeaveOneOutSplitter._get_unique_param_values(dataset, param, min_count)
        input_params = LeaveOneOutSplitter._update_split_count(input_params, unique_values)
        train_indices, test_indices = LeaveOneOutSplitter._get_train_test_indices(dataset, unique_values, param)

        train_datasets, test_datasets = LeaveOneOutSplitter._make_datasets_from_indices(unique_values, dataset, train_indices, test_indices,
                                                                                        input_params)

        return train_datasets, test_datasets

    @staticmethod
    def _update_split_count(input_params: DataSplitterParams, unique_values):

        input_params.split_config.split_count = unique_values.shape[0]
        input_params.split_count = input_params.split_config.split_count

        return input_params

    @staticmethod
    def _make_datasets_from_indices(unique_values, dataset, train_indices, test_indices, input_params):
        train_datasets, test_datasets = [], []
        for index, value in enumerate(unique_values):
            train_datasets.append(Util.make_dataset(dataset, train_indices[value], input_params, index, type(dataset).TRAIN))
            test_datasets.append(Util.make_dataset(dataset, test_indices[value], input_params, index, type(dataset).TEST))

        return train_datasets, test_datasets

    @staticmethod
    def _get_unique_param_values(dataset, param, min_count):
        if isinstance(dataset, ReceptorDataset):
            parameter_values = [receptor.metadata[param] for receptor in dataset.get_data()]
        elif isinstance(dataset, SequenceDataset):
            parameter_values = [seq.metadata.custom_params[param] for seq in dataset.get_data()]
        else:
            raise RuntimeError(f"{LeaveOneOutSplitter.__name__}: dataset of type {type(dataset)} cannot be used with this splitter.")

        unique_values, count = np.unique(parameter_values, return_counts=True)

        assert all(el > min_count for el in count), f"DataSplitter: there are not enough examples with different values of the parameter {param} " \
                                                    f"to split the dataset."

        return unique_values

    @staticmethod
    def _get_train_test_indices(dataset, unique_values, param):
        train_indices, test_indices = {value: [] for value in unique_values}, {value: [] for value in unique_values}
        for index, receptor in enumerate(dataset.get_data()):
            for value in unique_values:
                if LeaveOneOutSplitter._get_key_from_element_obj(receptor, param) == value:
                    test_indices[value].append(index)
                else:
                    train_indices[value].append(index)

        return train_indices, test_indices

    @staticmethod
    def _get_key_from_element_obj(obj, param):
        if isinstance(obj, ReceptorSequence):
            return obj.metadata.custom_params[param]
        else:
            return obj.metadata[param]



import logging
import os
from pathlib import Path

from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.dsl.definition_parsers.DefinitionParserOutput import DefinitionParserOutput
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.hyperparameter_optimization.config.ReportConfig import ReportConfig
from immuneML.hyperparameter_optimization.config.SplitConfig import SplitConfig
from immuneML.util.Logger import log
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.ReflectionHandler import ReflectionHandler
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.instructions.TrainMLModelInstruction import TrainMLModelInstruction
from scripts.DocumentatonFormat import DocumentationFormat
from scripts.specification_util import write_class_docs


class InstructionParser:

    keyword = "instructions"

    @staticmethod
    def parse(definition_output: DefinitionParserOutput, path):

        specification = definition_output.specification
        symbol_table = definition_output.symbol_table

        if InstructionParser.keyword in specification:
            if len(specification[InstructionParser.keyword].keys()) > 1:
                logging.warning(f"InstructionParser: multiple instructions were listed in the specification (under keys "
                                f"{str(list(specification[InstructionParser.keyword].keys()))[1:-1]}). "
                                "These instructions are independent and results from one instruction are not available to others. "
                                "If this is the intended behavior, please ignore this warning. If the output of one instruction is needed for the "
                                "other, please use separate YAML specifications and separate runs to perform the analysis.")

            for key in specification[InstructionParser.keyword]:
                assert type(specification[InstructionParser.keyword][key]) == dict, f"InstructionParser: instructions are incorrectly defined. Please make sure each instruction is defined under a unique keyword, like this:\n" \
                                                                                    f"instructions:\n" \
                                                                                    f"  my_instruction:\n" \
                                                                                    f"    type: InstructionType\n" \
                                                                                    f"    ... other parameters\n\n" \
                                                                                    f"Found keyword '{key}' containing '{specification[InstructionParser.keyword][key]}', which is a {type(specification[InstructionParser.keyword][key]).__name__}, " \
                                                                                    f"expected a dictionary instead."

                specification[InstructionParser.keyword][key], symbol_table = \
                    InstructionParser.parse_instruction(key, specification[InstructionParser.keyword][key], symbol_table, path)
        else:
            specification[InstructionParser.keyword] = {}

        return symbol_table, specification[InstructionParser.keyword]

    @staticmethod
    @log
    def parse_instruction(key: str, instruction: dict, symbol_table: SymbolTable, path) -> tuple:

        ParameterValidator.assert_keys_present(list(instruction.keys()), ["type"], InstructionParser.__name__, key)
        valid_instructions = [cls[:-6] for cls in ReflectionHandler.discover_classes_by_partial_name("Parser", "dsl/instruction_parsers/")]
        ParameterValidator.assert_in_valid_list(instruction["type"], valid_instructions, "InstructionParser", "type")

        default_params = DefaultParamsLoader.load("instructions/", instruction["type"])
        instruction = {**default_params, **instruction}
        parser = ReflectionHandler.get_class_by_name("{}Parser".format(instruction["type"]), "instruction_parsers/")()
        instruction_object = parser.parse(key, instruction, symbol_table, path)

        symbol_table.add(key, SymbolType.INSTRUCTION, instruction_object)
        return instruction, symbol_table

    @staticmethod
    def generate_docs(path: Path):
        inst_path = PathBuilder.build(path / "instructions")
        instructions = sorted(ReflectionHandler.all_nonabstract_subclasses(Instruction, "Instruction", subdirectory='instructions/'), key=lambda x: x.__name__)

        inst_paths = {}

        for instruction in instructions:
            instruction_name = instruction.__name__[:-11]
            if hasattr(InstructionParser, f"make_{instruction_name.lower()}_docs"):
                fn = getattr(InstructionParser, f"make_{instruction_name.lower()}_docs")
                file_path = fn(inst_path)
            else:
                file_path = InstructionParser.make_docs(instruction, instruction_name, inst_path)

            inst_paths[instruction_name] = file_path

        inst_file_path = inst_path / "instructions.rst"
        with inst_file_path.open('w') as file:
            for key, item in inst_paths.items():
                lines = f"{key}\n---------------------------\n.. include:: {os.path.relpath(item, EnvironmentSettings.source_docs_path)}\n"
                file.writelines(lines)

    @staticmethod
    def make_docs(instruction, name, path: Path):
        file_path = path / f"{name}.rst"

        with file_path.open("w") as file:
            write_class_docs(DocumentationFormat(instruction, "", DocumentationFormat.LEVELS[1]), file)
        return file_path



from dataclasses import dataclass, field
from typing import List

from immuneML.reports.ReportOutput import ReportOutput


@dataclass
class ReportResult:
    name: str = None
    info: str = field(default_factory=lambda: '')  # optional extra info about this report to display to the user
    output_figures: List[ReportOutput] = field(default_factory=lambda: [])
    output_tables: List[ReportOutput] = field(default_factory=lambda: [])
    output_text: List[ReportOutput] = field(default_factory=lambda: [])




from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.presentation.PresentationFormat import PresentationFormat
from immuneML.presentation.html.ClusteringHTMLBuilder import ClusteringHTMLBuilder
from immuneML.presentation.html.DatasetExportHTMLBuilder import DatasetExportHTMLBuilder
from immuneML.presentation.html.ExploratoryAnalysisHTMLBuilder import ExploratoryAnalysisHTMLBuilder
from immuneML.presentation.html.FeasibilitySummaryHTMLBuilder import FeasibilitySummaryHTMLBuilder
from immuneML.presentation.html.HPHTMLBuilder import HPHTMLBuilder
from immuneML.presentation.html.LIgOSimulationHTMLBuilder import LIgOSimulationHTMLBuilder
from immuneML.presentation.html.MLApplicationHTMLBuilder import MLApplicationHTMLBuilder
from immuneML.presentation.html.SubsamplingHTMLBuilder import SubsamplingHTMLBuilder
from immuneML.presentation.html.GenModelHTMLBuilder import GenModelHTMLBuilder
from immuneML.simulation.LigoSimState import LigoSimState
from immuneML.workflows.instructions.clustering.ClusteringInstruction import ClusteringState
from immuneML.workflows.instructions.dataset_generation.DatasetExportState import DatasetExportState
from immuneML.workflows.instructions.exploratory_analysis.ExploratoryAnalysisState import ExploratoryAnalysisState
from immuneML.workflows.instructions.ligo_sim_feasibility.FeasibilitySummaryInstruction import FeasibilitySummaryState
from immuneML.workflows.instructions.ml_model_application.MLApplicationState import MLApplicationState
from immuneML.workflows.instructions.subsampling.SubsamplingState import SubsamplingState
from immuneML.workflows.instructions.train_gen_model.TrainGenModelInstruction import GenModelState, TrainGenModelState


class PresentationFactory:

    @staticmethod
    def make_presentation_builder(state, presentation_format: PresentationFormat):
        if isinstance(state, TrainMLModelState) and presentation_format == PresentationFormat.HTML:
            return HPHTMLBuilder
        elif isinstance(state, ExploratoryAnalysisState) and presentation_format == PresentationFormat.HTML:
            return ExploratoryAnalysisHTMLBuilder
        elif isinstance(state, DatasetExportState) and presentation_format == PresentationFormat.HTML:
            return DatasetExportHTMLBuilder
        elif isinstance(state, MLApplicationState) and presentation_format == PresentationFormat.HTML:
            return MLApplicationHTMLBuilder
        elif isinstance(state, SubsamplingState) and presentation_format == PresentationFormat.HTML:
            return SubsamplingHTMLBuilder
        elif isinstance(state, LigoSimState) and presentation_format == PresentationFormat.HTML:
            return LIgOSimulationHTMLBuilder
        elif isinstance(state, FeasibilitySummaryState) and presentation_format == PresentationFormat.HTML:
            return FeasibilitySummaryHTMLBuilder
        elif (isinstance(state, TrainGenModelState) or isinstance(state, GenModelState)) and presentation_format == PresentationFormat.HTML:
            return GenModelHTMLBuilder
        elif isinstance(state, ClusteringState) and presentation_format == PresentationFormat.HTML:
            return ClusteringHTMLBuilder
        else:
            raise ValueError(f"PresentationFactory: state and format combination ({type(state).__name__}, "
                             f"{presentation_format.name}) is not supported.")



from pathlib import Path

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.Label import Label
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.reports.Report import Report


class MLReport(Report):
    """
    ML model reports show some type of features or statistics about a single trained ML model.

    In the :ref:`TrainMLModel` instruction, ML model reports can be specified inside the 'selection' or 'assessment' specification under the key 'reports/models'.
    Example:

    .. indent with spaces
    .. code-block:: yaml

        my_instruction:
            type: TrainMLModel
            selection:
                reports:
                    models:
                        - my_ml_report
                # other parameters...
            assessment:
                reports:
                    models:
                        - my_ml_report
                # other parameters...
            # other parameters...
    """
    DOCS_TITLE = "ML model reports"

    def __init__(self, train_dataset: Dataset = None, test_dataset: Dataset = None, method: MLMethod = None,
                 result_path: Path = None, name: str = None, hp_setting: HPSetting = None, label: Label =None, number_of_processes: int = 1):
        """
        The arguments defined below are set at runtime by the instruction.
        Concrete classes inheriting MLReport may include additional parameters that will be set by the user in the form of input arguments.

        train_dataset (Dataset): a dataset object (repertoire, receptor or sequence dataset) with encoded_data attribute set to an EncodedData object that was used for training the ML method
        test_dataset (Dataset): same as train_dataset, except it is not used for training and then maybe be used for testing the method
        method (MLMethod): a trained instance of a concrete subclass of MLMethod object
        result_path (Path): location where the report results will be stored
        hp_setting (HPSetting): a HPSetting object describing the ML method, encoding and preprocessing used
        label (Label): the label for which the model was trained
        name (str): user-defined name of the report used in the HTML overview automatically generated by the platform
        number_of_processes (int): how many processes should be created at once to speed up the analysis. For personal machines, 4 or 8 is usually a good choice.
        """
        super().__init__(name=name, result_path=result_path, number_of_processes=number_of_processes)
        self.train_dataset = train_dataset
        self.test_dataset = test_dataset
        self.method = method
        self.name = name
        self.hp_setting = hp_setting
        self.label = label

